"""Application settings."""

from __future__ import annotations

from logging import getLogger
from typing import Any, TypedDict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from slack_bolt import App
from typing_extensions import NotRequired

from django_slack_tools.slack_messages.backends.base import BaseBackend

APP_SETTINGS_KEY = "DJANGO_SLACK_TOOLS"
"Django settings key for this application."

logger = getLogger(__name__)


# TODO(lasuillard): Configuration getting dirty, need refactoring
class AppSettings:
    """Application settings."""

    backend: BaseBackend

    def __init__(self, settings_dict: ConfigDict | None = None) -> None:
        """Initialize app settings.

        Args:
            settings_dict: Settings dictionary. If `None`, try to load from Django settings (const `APP_SETTINGS_KEY`).
                Defaults to `None`.

        Raises:
            ImproperlyConfigured: Required configuration not provided or is unexpected.
        """
        if not settings_dict:
            settings_dict = getattr(settings, APP_SETTINGS_KEY, None)

        if settings_dict is None:
            msg = "Neither `settings_dict` provided or `django_slack_tools` settings found in Django settings."
            raise ImproperlyConfigured(msg)

        # Slack app
        slack_app = settings_dict["SLACK_APP"]
        if isinstance(slack_app, str):
            slack_app = import_string(slack_app)

        if callable(slack_app):
            slack_app = slack_app()

        if not isinstance(slack_app, App):
            msg = "Provided `SLACK_APP` config is not Slack app."
            raise ImproperlyConfigured(msg)

        self._slack_app = slack_app

        # Find backend class
        messaging_backend = import_string(settings_dict["BACKEND"]["NAME"])
        if not issubclass(messaging_backend, BaseBackend):
            msg = "Provided backend is not a subclass of `{qualified_path}` class.".format(
                qualified_path=f"{BaseBackend.__module__}.{BaseBackend.__name__}",
            )
            raise ImproperlyConfigured(msg)

        # Initialize with provided options
        self.backend = messaging_backend(**settings_dict["BACKEND"]["OPTIONS"])

        # Message delivery default
        self.default_policy_code = settings_dict.get("DEFAULT_POLICY_CODE", "DEFAULT")

        # Lazy policy defaults
        self.lazy_policy_enabled = settings_dict.get("LAZY_POLICY_ENABLED", False)
        self.default_template = settings_dict.get(
            "DEFAULT_POLICY_CODE",
            {"text": "No template configured for lazily created policy {policy}"},
        )
        self.default_recipient = settings_dict.get("DEFAULT_RECIPIENT", "DEFAULT")

    @property
    def slack_app(self) -> App:
        """Registered Slack app or `None`."""
        return self._slack_app


app_settings = AppSettings()


class ConfigDict(TypedDict):
    """Root config dict."""

    SLACK_APP: App | str | None
    "Import string to Slack app. Used for workspace-related stuffs."

    BACKEND: BackendConfig
    "Nested backend config."

    LAZY_POLICY_ENABLED: NotRequired[bool]
    "Whether to enable lazy policy by default."

    DEFAULT_POLICY_CODE: NotRequired[str]
    "Default policy code used when sending messages via policy with no policy specified."

    DEFAULT_TEMPLATE: NotRequired[Any]
    "Default template for lazy policy."

    DEFAULT_RECIPIENT: NotRequired[str]
    "Default recipient alias for lazy policy."


class BackendConfig(TypedDict):
    """Backend config dict."""

    NAME: str
    "Import path to backend class."

    OPTIONS: dict[str, Any]
    "Options to pass to backend class on initialization."



from .base import BaseBackend
from .dummy import DummyBackend
from .logging import LoggingBackend
from .slack import SlackBackend, SlackRedirectBackend

__all__ = (
    "BaseBackend",
    "SlackBackend",
    "SlackRedirectBackend",
    "DummyBackend",
    "LoggingBackend",
)



from django_slack_tools.utils.widgets import JSONWidget


class TestJSONWidget:
    def test_format_value(self) -> None:
        widget = JSONWidget()

        # Pretty format with some indents
        assert (
            widget.format_value('{"foo": "bar"}')
            == """{
  "foo": "bar"
}"""
        )

        # If erroneous, just return value
        assert widget.format_value("}not_valid_json") == "}not_valid_json"



from django_slack_tools.slack_messages.models import SlackMessagingPolicy
from tests._helpers import ModelTestBase

from ._factories import SlackMessageRecipientFactory, SlackMessagingPolicyFactory


class TestSlackMessagingPolicy(ModelTestBase):
    model_cls = SlackMessagingPolicy
    factory_cls = SlackMessagingPolicyFactory

    def test_str(self) -> None:
        # Enabled with recipients
        recipients = SlackMessageRecipientFactory.create_batch(size=3)
        instance = self.factory_cls.create(code="OP-641-XA01", enabled=True, recipients=recipients)
        assert str(instance) == "OP-641-XA01 (enabled, 3 recipients)"

        # Disabled but existing recipients
        recipients = SlackMessageRecipientFactory.create_batch(size=1)
        instance = self.factory_cls.create(code="REPORT-1178-CR9217", enabled=False, recipients=recipients)
        assert str(instance) == "REPORT-1178-CR9217 (disabled, 1 recipients)"



from django.contrib import admin
from django.urls import path

from django_slack_tools.views import SlackEventHandlerView
from todo.slack_app import app

urlpatterns = [
    path("admin/", admin.site.urls),
    path("slack/events/", SlackEventHandlerView.as_view(app=app)),
]



"""Django model mixins."""

from __future__ import annotations

from django.db import models
from django.utils.translation import gettext_lazy as _


class TimestampMixin(models.Model):
    """Django model mixin for created and modified timestamp fields."""

    created = models.DateTimeField(
        verbose_name=_("Created"),
        help_text=_("When instance created."),
        auto_now_add=True,
    )
    last_modified = models.DateTimeField(
        verbose_name=_("Last Modified"),
        help_text=_("When instance modified recently."),
        auto_now=True,
    )

    class Meta:  # noqa: D106
        abstract = True



"""Messaging policy model."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_slack_tools.utils.model_mixins import TimestampMixin
from django_slack_tools.utils.slack import header_validator

from .message_recipient import SlackMessageRecipient

if TYPE_CHECKING:
    from typing import Any


class SlackMessagingPolicyManager(models.Manager["SlackMessagingPolicy"]):
    """Manager for Slack messaging policies."""


class SlackMessagingPolicy(TimestampMixin, models.Model):
    """An Slack messaging policy which determines message content and those who receive messages."""

    class TemplateType(models.TextChoices):
        """Possible template types."""

        DICT = "D", _("Dictionary")
        "Dictionary-based template."

        DJANGO = "DJ", _("Django")
        "Django XML-based template."

        DJANGO_INLINE = "DI", _("Django Inline")
        "Django inline template."

        UNKNOWN = "?", _("Unknown")
        "Unknown template type."

    code = models.CharField(
        verbose_name=_("Code"),
        help_text=_("Unique message code for lookup, mostly by human."),
        max_length=32,
        unique=True,
    )
    enabled = models.BooleanField(
        verbose_name=_("Enabled"),
        help_text=_("Turn on or off current messaging policy."),
        default=True,
    )
    recipients = models.ManyToManyField(
        SlackMessageRecipient,
        verbose_name=_("Message recipients"),
        help_text=_("Those who will receive messages."),
    )
    header_defaults = models.JSONField(
        verbose_name=_("Default header"),
        help_text=_("Default header values applied to messages on creation."),
        validators=[header_validator],
        blank=True,
        default=dict,
    )
    template_type = models.CharField(
        verbose_name=_("Template type"),
        help_text=_("Type of message template."),
        max_length=2,
        choices=TemplateType.choices,
        default=TemplateType.DICT,
    )
    template: models.JSONField[Any] = models.JSONField(
        verbose_name=_("Message template object"),
        help_text=_("Dictionary-based template object."),
        null=True,
        blank=True,
    )

    # Type is too obvious but due to limits...
    objects: SlackMessagingPolicyManager = SlackMessagingPolicyManager()

    class Meta:  # noqa: D106
        verbose_name = _("Messaging Policy")
        verbose_name_plural = _("Messaging Policies")

    def __str__(self) -> str:
        num_recipients = self.recipients.all().count()
        if self.enabled:
            return _("{code} (enabled, {num_recipients} recipients)").format(
                code=self.code,
                num_recipients=num_recipients,
            )

        return _("{code} (disabled, {num_recipients} recipients)").format(code=self.code, num_recipients=num_recipients)



"""Miscellaneous features."""

from __future__ import annotations

import json
import urllib.parse


def get_block_kit_builder_url(*, team_id: str = "", blocks: list | None = None, attachments: list | None = None) -> str:
    """Returns URL to Slack Block Kit Builder.

    Args:
        team_id: Slack team ID. Can be omitted, hoping the browser to redirect user who clicked the link optimistically.
        blocks: Message blocks. Defaults to None.
        attachments: Message attachments. Defaults to None.

    Raises:
        ValueError: Thrown if not only single value of blocks or attachments provided.

    Returns:
        URL to Block Kit Builder preview.
    """
    if (not blocks and not attachments) or (blocks and attachments):
        msg = "Only one of `blocks` or `attachments` should be provided."
        raise ValueError(msg)

    payload = {"blocks": blocks} if blocks else {"attachments": attachments} if attachments else {}
    payload_urlencoded = urllib.parse.quote(json.dumps(payload))
    return f"https://app.slack.com/block-kit-builder/{team_id}#{payload_urlencoded}"



from __future__ import annotations

from typing import Generator
from unittest import mock

import pytest
from slack_bolt import App


@pytest.fixture(scope="session")
def slack_app() -> App:
    """Dummy Slack app fixture. It won't work."""
    return App(
        token="stupid-sandwich",  # noqa: S106
        signing_secret="peanut-butter",  # noqa: S106
        token_verification_enabled=False,
    )


@pytest.fixture
def mock_slack_client() -> Generator[mock.Mock, None, None]:
    """Mock `slack_bolt.App.client`."""
    with mock.patch("slack_bolt.App.client") as m:
        yield m



from django_slack_tools.slack_messages.models import SlackMessageRecipient
from tests._helpers import ModelTestBase

from ._factories import SlackMentionFactory, SlackMessageRecipientFactory


class TestSlackMessageRecipient(ModelTestBase):
    model_cls = SlackMessageRecipient
    factory_cls = SlackMessageRecipientFactory

    def test_str(self) -> None:
        instance = self.factory_cls.create(
            alias="Robots in the bar",
            channel="#robots-in-the-bar",
            mentions=SlackMentionFactory.create_batch(size=3),
        )
        assert str(instance) == "Robots in the bar (#robots-in-the-bar, 3 mentions)"



"""Message recipients model."""

from __future__ import annotations

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_slack_tools.utils.model_mixins import TimestampMixin

from .mention import SlackMention


class SlackMessageRecipientManager(models.Manager["SlackMessageRecipient"]):
    """Manager for message recipients model."""


class SlackMessageRecipient(TimestampMixin, models.Model):
    """People or group in channels receive messages."""

    alias = models.CharField(
        verbose_name=_("Alias"),
        help_text=_("Alias for this recipient."),
        max_length=256,
        unique=True,
    )
    channel = models.CharField(
        verbose_name=_("Channel"),
        help_text=_("Slack channel ID where messages will be sent."),
        max_length=128,
        blank=False,
    )
    channel_name = models.CharField(
        verbose_name=_("Channel name"),
        help_text=_("Display name of channel."),
        max_length=256,
        blank=True,
        default="",
    )
    mentions = models.ManyToManyField(
        SlackMention,
        verbose_name=_("Mentions"),
        help_text=_("List of mentions."),
        blank=True,
    )

    objects: SlackMessageRecipientManager = SlackMessageRecipientManager()

    class Meta:  # noqa: D106
        verbose_name = _("Recipient")
        verbose_name_plural = _("Recipients")

    def __str__(self) -> str:
        num_mentions = self.mentions.count()

        return _("{alias} ({channel}, {num_mentions} mentions)").format(
            alias=self.alias,
            channel=self.channel,
            num_mentions=num_mentions,
        )



# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from .base import BaseTemplate

if TYPE_CHECKING:
    from typing import Any


import logging

logger = logging.getLogger(__name__)


class DictTemplate(BaseTemplate):
    """Simple dictionary-based template."""

    def __init__(self, template: dict) -> None:
        """Initialize template.

        Args:
            template: Dictionary template.
            kwargs: Keyword arguments passed to template.
        """
        self.template = template

    def render(self, *, context: dict[str, Any] | None = None) -> dict:  # noqa: D102
        context = {} if context is None else context
        result = self.template.copy()
        for k, v in result.items():
            result[k] = _format_obj(v, context=context)

        return result


T = TypeVar("T", dict, list, str)


def _format_obj(obj: T, *, context: dict[str, Any]) -> T:
    """Format object recursively."""
    if isinstance(obj, dict):
        return {k: _format_obj(v, context=context) for k, v in obj.items()}

    if isinstance(obj, str):
        return obj.format_map(context)

    if isinstance(obj, list):
        return [_format_obj(item, context=context) for item in obj]

    return obj



# Generated by Django 4.2.9 on 2024-01-16 13:32

import django.db.models.deletion
from django.db import migrations, models

import django_slack_tools.utils.slack.django


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SlackMention",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, help_text="When instance created.", verbose_name="Created"),
                ),
                (
                    "last_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When instance modified recently.",
                        verbose_name="Last Modified",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[("U", "User"), ("G", "Group"), ("S", "Special"), ("?", "Unknown")],
                        help_text="Type of mentions.",
                        max_length=1,
                        verbose_name="Type",
                    ),
                ),
                (
                    "name",
                    models.CharField(help_text="Human-friendly mention name.", max_length=128, verbose_name="Name"),
                ),
                (
                    "mention_id",
                    models.CharField(
                        help_text="User or group ID, or raw mention itself.",
                        max_length=32,
                        verbose_name="Mention ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "Mention",
                "verbose_name_plural": "Mentions",
            },
        ),
        migrations.CreateModel(
            name="SlackMessageRecipient",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, help_text="When instance created.", verbose_name="Created"),
                ),
                (
                    "last_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When instance modified recently.",
                        verbose_name="Last Modified",
                    ),
                ),
                (
                    "alias",
                    models.CharField(
                        help_text="Alias for this recipient.",
                        max_length=256,
                        unique=True,
                        verbose_name="Alias",
                    ),
                ),
                (
                    "channel",
                    models.CharField(
                        help_text="Slack channel ID where messages will be sent.",
                        max_length=128,
                        verbose_name="Channel",
                    ),
                ),
                (
                    "channel_name",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Display name of channel.",
                        max_length=256,
                        verbose_name="Channel name",
                    ),
                ),
                (
                    "mentions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="List of mentions.",
                        to="slack_messages.slackmention",
                        verbose_name="Mentions",
                    ),
                ),
            ],
            options={
                "verbose_name": "Recipient",
                "verbose_name_plural": "Recipients",
            },
        ),
        migrations.CreateModel(
            name="SlackMessagingPolicy",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, help_text="When instance created.", verbose_name="Created"),
                ),
                (
                    "last_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When instance modified recently.",
                        verbose_name="Last Modified",
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        help_text="Unique message code for lookup, mostly by human.",
                        max_length=32,
                        unique=True,
                        verbose_name="Code",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Turn on or off current messaging policy.",
                        verbose_name="Enabled",
                    ),
                ),
                (
                    "header_defaults",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Default header values applied to messages on creation.",
                        validators=[django_slack_tools.utils.slack.django.header_validator],
                        verbose_name="Default header",
                    ),
                ),
                (
                    "template",
                    models.JSONField(
                        blank=True,
                        help_text="Dictionary-based template object.",
                        null=True,
                        verbose_name="Message template object",
                    ),
                ),
                (
                    "recipients",
                    models.ManyToManyField(
                        help_text="Those who will receive messages.",
                        to="slack_messages.slackmessagerecipient",
                        verbose_name="Message recipients",
                    ),
                ),
            ],
            options={
                "verbose_name": "Messaging Policy",
                "verbose_name_plural": "Messaging Policies",
            },
        ),
        migrations.CreateModel(
            name="SlackMessage",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, help_text="When instance created.", verbose_name="Created"),
                ),
                (
                    "last_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When instance modified recently.",
                        verbose_name="Last Modified",
                    ),
                ),
                (
                    "channel",
                    models.CharField(
                        help_text="ID of channel this message sent to.",
                        max_length=128,
                        verbose_name="Channel",
                    ),
                ),
                (
                    "header",
                    models.JSONField(
                        help_text="Slack control arguments. Allowed fields are `mrkdwn`, `parse`, `reply_broadcast`, `thread_ts`, `unfurl_links`, `unfurl_media`.",  # noqa: E501
                        validators=[django_slack_tools.utils.slack.django.header_validator],
                        verbose_name="Header",
                    ),
                ),
                (
                    "body",
                    models.JSONField(
                        help_text="Message body. Allowed fields are `attachments`, `body`, `text`, `icon_emoji`, `icon_url`, `metadata`, `username`.",  # noqa: E501
                        validators=[django_slack_tools.utils.slack.django.body_validator],
                        verbose_name="Body",
                    ),
                ),
                (
                    "ok",
                    models.BooleanField(
                        default=None,
                        help_text="Whether Slack API respond with OK. If never sent, will be `null`.",
                        null=True,
                        verbose_name="OK",
                    ),
                ),
                (
                    "permalink",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Permanent link for this message.",
                        max_length=256,
                        verbose_name="Permalink",
                    ),
                ),
                (
                    "ts",
                    models.CharField(
                        blank=True,
                        help_text="ID of an Slack message.",
                        max_length=32,
                        null=True,
                        unique=True,
                        verbose_name="Message ID",
                    ),
                ),
                (
                    "parent_ts",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="ID of current conversation thread.",
                        max_length=32,
                        verbose_name="Thread ID",
                    ),
                ),
                (
                    "request",
                    models.JSONField(
                        blank=True,
                        help_text="Dump of request content for debugging.",
                        null=True,
                        verbose_name="Request",
                    ),
                ),
                (
                    "response",
                    models.JSONField(
                        blank=True,
                        help_text="Dump of response content for debugging.",
                        null=True,
                        verbose_name="Response",
                    ),
                ),
                (
                    "exception",
                    models.TextField(blank=True, help_text="Exception message if any.", verbose_name="Exception"),
                ),
                (
                    "policy",
                    models.ForeignKey(
                        blank=True,
                        help_text="Messaging policy for this message.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="slack_messages.slackmessagingpolicy",
                        verbose_name="Messaging Policy",
                    ),
                ),
            ],
            options={
                "verbose_name": "Message",
                "verbose_name_plural": "Messages",
                "ordering": ("-created",),
            },
        ),
    ]



from django.contrib import admin
from django.urls import include, path

from django_slack_tools.views import SlackEventHandlerView
from testproj.config.slack_app import app

urlpatterns = [
    path("admin/", admin.site.urls),
    path("slack/events", SlackEventHandlerView.as_view(app=app), name="slack_events"),
    path("__debug__/", include("debug_toolbar.urls")),
]



import os

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "todo.settings")

from celery import Celery
from celery.schedules import crontab
from django.conf import settings

app = Celery("todo")

app.conf.broker_url = "redis://localhost:6379/0"

app.conf.result_backend = "django-db"

if settings.USE_TZ:
    app.conf.timezone = settings.TIME_ZONE

app.conf.beat_scheduler = "django_celery_beat.schedulers:DatabaseScheduler"
app.conf.beat_schedule = {
    "Cleanup old Slack messages": {
        "task": "django_slack_tools.slack_messages.tasks.cleanup_old_messages",
        "schedule": crontab(minute=0, hour=4),  # Every 4 AM
        "kwargs": {
            "threshold_seconds": 1 * 24 * 60 * 60,  # Retain for a day
        },
    },
}

app.autodiscover_tasks()



from django.apps import AppConfig


class TodoAppConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "todo_app"

    def ready(self) -> None:  # noqa: D102
        from . import signals, slack_listeners  # noqa: F401



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable
from unittest import mock

import pytest

from django_slack_tools.slack_messages.admin import SlackMessageRecipientAdmin
from django_slack_tools.slack_messages.admin.message_recipient import _get_channels
from django_slack_tools.slack_messages.models import SlackMessageRecipient
from tests._factories import SlackResponseFactory
from tests._helpers import ModelAdminTestBase
from tests.slack_messages.models._factories import SlackMessageRecipientFactory

if TYPE_CHECKING:
    from django.test import Client
    from django.test.client import _MonkeyPatchedWSGIResponse


_channels = {
    "CHANNEL001": "channel-001",
    "CHANNEL002": "channel-002",
}


class TestSlackMessageRecipientAdmin(ModelAdminTestBase):
    admin_cls = SlackMessageRecipientAdmin
    model_cls = SlackMessageRecipient
    factory_cls = SlackMessageRecipientFactory

    pytestmark = pytest.mark.django_db()

    def _update_channel_names(
        self,
        *,
        client: Client,
        ids: Iterable[SlackMessageRecipient],
    ) -> _MonkeyPatchedWSGIResponse:
        return client.post(
            self._reverse("changelist"),
            {
                "action": "_update_channel_names",
                "_selected_action": ids,
            },
        )

    def test_update_channel_names(self, admin_client: Client) -> None:
        # These recipients should update
        recipients_to_update = [
            self.factory_cls.create(channel="CHANNEL001"),
            self.factory_cls.create(channel="CHANNEL002"),
        ]
        ids = [r.id for r in recipients_to_update]

        # Perform action
        with mock.patch(
            "django_slack_tools.slack_messages.admin.message_recipient._get_channels",
            return_value=_channels,
        ):
            response = self._update_channel_names(client=admin_client, ids=ids)

        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == ["Updated 2 recipients."]

        # Check changes
        recipients = SlackMessageRecipient.objects.filter(id__in=ids)
        assert list(recipients.values("channel", "channel_name")) == [
            {"channel": "CHANNEL001", "channel_name": "#channel-001"},
            {"channel": "CHANNEL002", "channel_name": "#channel-002"},
        ]

    def test_update_channel_names_no_match_for_some(self, admin_client: Client) -> None:
        # These recipients should update
        recipients_to_update = [
            self.factory_cls.create(channel="CHANNEL001"),
            self.factory_cls.create(channel="CHANNEL002"),
            # No match for this
            self.factory_cls.create(channel="CHANNEL003", channel_name="no-channel-for-this"),
        ]
        ids = [r.id for r in recipients_to_update]

        # Perform action
        with mock.patch(
            "django_slack_tools.slack_messages.admin.message_recipient._get_channels",
            return_value=_channels,
        ):
            response = self._update_channel_names(client=admin_client, ids=ids)

        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == [
            "Updated 2 recipients successfully and there were 1 recipients failed to update because no matching data.",
        ]

        # Check changes
        recipients = SlackMessageRecipient.objects.filter(id__in=ids)
        assert list(recipients.values("channel", "channel_name")) == [
            {"channel": "CHANNEL001", "channel_name": "#channel-001"},
            {"channel": "CHANNEL002", "channel_name": "#channel-002"},
            {"channel": "CHANNEL003", "channel_name": "no-channel-for-this"},
        ]


def test_get_channels(mock_slack_client: mock.Mock) -> None:
    mock_slack_client.conversations_list.return_value = SlackResponseFactory(
        data={
            "ok": True,
            "channels": [
                {"id": "CHANNEL001", "name": "channel-001"},
                {"id": "CHANNEL002", "name": "channel-002"},
            ],
        },
    )
    channels = _get_channels()

    assert channels == {"CHANNEL001": "channel-001", "CHANNEL002": "channel-002"}



# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from django.contrib import admin
from django.contrib.admin.filters import DateFieldListFilter
from django.db import models
from django.db.models import Count
from django.db.models.query import QuerySet
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from django_slack_tools.slack_messages.models import SlackMessagingPolicy
from django_slack_tools.utils.slack import get_block_kit_builder_url
from django_slack_tools.utils.widgets import JSONWidget

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext import StrOrPromise

    # I can't make any better idea for my experience, extending model instance with annotates
    # If you, reader, knows better one, plz make PR :)
    # NOTE: Pylance complains `django_stubs_ext.WithAnnotations` about its call signature
    class SlackMessagingPolicyWithAnnotates(SlackMessagingPolicy):  # noqa: D101
        num_recipients: int

else:
    SlackMessagingPolicyWithAnnotates = SlackMessagingPolicy


@admin.register(SlackMessagingPolicy)
class SlackMessagingPolicyAdmin(admin.ModelAdmin):
    """Admin for messaging policies."""

    def get_queryset(self, request: HttpRequest) -> QuerySet[SlackMessagingPolicyWithAnnotates]:  # noqa: D102
        return cast(
            QuerySet["SlackMessagingPolicyWithAnnotates"],  # Unsafe force type casting
            super()
            .get_queryset(request)
            .annotate(
                # Avoid calling `.recipients.count()` per records
                num_recipients=Count("recipients"),
            ),
        )

    readonly_fields = (
        "id",
        "_count_recipients",
        "_blocks_block_kit_builder_url",
        "_attachments_block_kit_builder_url",
        "created",
        "last_modified",
    )

    @admin.display(description=_("Number of Recipients"))
    def _count_recipients(self, instance: SlackMessagingPolicyWithAnnotates) -> int:
        return instance.num_recipients

    @admin.display(description=_("Blocks Preview"))
    def _blocks_block_kit_builder_url(self, instance: SlackMessagingPolicy) -> StrOrPromise:
        """Generate shortcut URL to Slack Block Kit Builder page for current policy template."""
        template = instance.template
        if not instance.template:
            return _("Template is empty, no link available.")

        if "blocks" not in template:
            return _("Template has no blocks.")

        url = get_block_kit_builder_url(blocks=template["blocks"])
        return format_html("<a href='{url}'>Link to Block Kit Builder</a>", url=url)

    @admin.display(description=_("Attachments Preview"))
    def _attachments_block_kit_builder_url(self, instance: SlackMessagingPolicy) -> StrOrPromise:
        """Generate shortcut URL to Slack Block Kit Builder page for current policy template."""
        template = instance.template
        if not instance.template:
            return _("Template is empty, no link available.")

        if "attachments" not in template:
            return _("Template has no attachments.")

        url = get_block_kit_builder_url(attachments=template["attachments"])
        return format_html("<a href='{url}'>Link to Block Kit Builder</a>", url=url)

    # Actions
    actions = ()

    # Changelist
    # ------------------------------------------------------------------------
    date_hierarchy = "last_modified"
    search_fields = ("code",)
    list_display = ("id", "code", "enabled", "_count_recipients", "created", "last_modified")
    list_display_links = ("id", "code")
    list_filter = (
        "enabled",
        ("created", DateFieldListFilter),
        ("last_modified", DateFieldListFilter),
    )

    # Change
    # ------------------------------------------------------------------------
    formfield_overrides = {  # noqa: RUF012
        models.JSONField: {"widget": JSONWidget},
    }
    fieldsets = (
        (
            None,
            {
                "fields": ("code", "enabled", "recipients", "header_defaults", "template"),
            },
        ),
        (
            _("Utility"),
            {
                "fields": ("_blocks_block_kit_builder_url", "_attachments_block_kit_builder_url"),
            },
        ),
        (
            _("Miscellaneous"),
            {
                "fields": ("id", "created", "last_modified"),
                "classes": ("collapse",),
            },
        ),
    )
    autocomplete_fields = ("recipients",)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
from django.contrib.messages import get_messages
from django.urls import reverse

if TYPE_CHECKING:
    from typing import Literal

    from django.contrib.admin import ModelAdmin
    from django.core.handlers.wsgi import WSGIRequest
    from django.db.models import Model
    from django.test.client import Client
    from factory.django import DjangoModelFactory


class ModelAdminTestBase:
    admin_cls: type[ModelAdmin]
    model_cls: type[Model]
    factory_cls: type[DjangoModelFactory]

    pytestmark = pytest.mark.django_db()

    def _reverse(
        self,
        view: Literal["changelist", "change", "delete", "history", "add"],
        *args: Any,
        **kwargs: Any,
    ) -> str:
        """Helper method returns reverse URL For admin view. e.g. `"/admin/django_slack_tools/slackmessage/add"`.

        Any additional arguments will be passed to `django.urls.reverse()`.
        """
        app_label = self.model_cls._meta.app_label
        model_name = self.model_cls._meta.model_name
        return reverse(f"admin:{app_label}_{model_name}_{view}", *args, **kwargs)

    def _get_messages(
        self,
        wsgi_request: WSGIRequest,
    ) -> list[str]:
        """Helper method to get all messages."""
        return [str(m) for m in get_messages(wsgi_request)]

    # NOTE: Below tests do some simple sanity checks only, extra test details should be provided by inherited classes

    def test_changelist(self, admin_client: Client) -> None:
        self.factory_cls.create_batch(size=3)
        url = self._reverse("changelist")

        # Test visit
        response = admin_client.get(url)
        assert response.status_code == 200

    def test_change(self, admin_client: Client) -> None:
        obj = self.factory_cls.create()
        url = self._reverse("change", kwargs={"object_id": obj.id})

        # Test visit
        response = admin_client.get(url)
        assert response.status_code == 200

    def test_delete(self, admin_client: Client) -> None:
        obj = self.factory_cls.create()
        url = self._reverse("delete", kwargs={"object_id": obj.id})

        # Visit object deletion page
        response = admin_client.get(url)
        assert response.status_code == 200

        # Delete object
        response = admin_client.delete(url)
        assert response.status_code == 200

    def test_history(self, admin_client: Client) -> None:
        obj = self.factory_cls.create()
        url = self._reverse("history", kwargs={"object_id": obj.id})

        # Test visit
        response = admin_client.get(url)
        assert response.status_code == 200

    def test_add(self, admin_client: Client) -> None:
        url = self._reverse("add")

        # Test visit
        response = admin_client.get(url)
        assert response.status_code == 200


class ModelTestBase:
    model_cls: type[Model]
    factory_cls: type[DjangoModelFactory]

    pytestmark = pytest.mark.django_db()

    # NOTE: Below tests do some simple sanity checks only, extra test details should be provided by inherited classes

    def test_instance_creation(self) -> None:
        obj = self.factory_cls.create()
        assert isinstance(obj, self.model_cls)



"""Some helpful views."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from slack_bolt.adapter.django import SlackRequestHandler

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse
    from slack_bolt import App


class SlackEventHandlerView(View):  # pragma: no cover; TODO(lasuillard): deal with this in future
    """View for handling Slack events."""

    app: App | None = None

    def __init__(self, *, app: App | Callable[[], App]) -> None:
        """Initialize view.

        Args:
            app: Slack app instance or callable which returns app.
        """
        if callable(app):
            app = app()

        self._event_handler = SlackRequestHandler(app=app)

        super().__init__()

    # Checking CSRF is nonsense because events come from Slack
    @method_decorator(csrf_exempt)
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:  # noqa: D102, ARG002
        return self._event_handler.handle(request)



"""Message recipients model."""

from __future__ import annotations

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_slack_tools.utils.model_mixins import TimestampMixin


class SlackMentionManager(models.Manager["SlackMention"]):
    """Manager for message recipients model."""


class SlackMention(TimestampMixin, models.Model):
    """People or group in channels receive messages."""

    class MentionType(models.TextChoices):
        """Possible mention types."""

        USER = "U", _("User")
        "User mentions. e.g. `@lasuillard`."

        GROUP = "G", _("Group")
        "Team mentions. e.g. `@backend`."

        SPECIAL = "S", _("Special")
        "Special mentions. e.g. `@here`, `@channel`, `@everyone`."

        UNKNOWN = "?", _("Unknown")
        "Unknown mention type."

    type = models.CharField(
        verbose_name=_("Type"),
        help_text=_("Type of mentions."),
        max_length=1,
        choices=MentionType.choices,
        blank=False,
    )
    name = models.CharField(
        verbose_name=_("Name"),
        help_text=_("Human-friendly mention name."),
        max_length=128,
    )
    mention_id = models.CharField(
        verbose_name=_("Mention ID"),
        help_text=_("User or group ID, or raw mention itself."),
        max_length=32,
    )

    objects: SlackMentionManager = SlackMentionManager()

    class Meta:  # noqa: D106
        verbose_name = _("Mention")
        verbose_name_plural = _("Mentions")

    def __str__(self) -> str:
        return _("{name} ({type}, {mention_id})").format(
            name=self.name,
            type=self.get_type_display(),
            mention_id=self.mention_id,
        )

    @property
    def mention(self) -> str:
        """Mention string for use in messages, e.g. `"<@{USER_ID}>"`."""
        if self.type == SlackMention.MentionType.USER:
            return f"<@{self.mention_id}>"

        if self.type == SlackMention.MentionType.GROUP:
            return f"<!subteam^{self.mention_id}>"

        return self.mention_id



from django.conf import settings
from slack_bolt import App

app = App(
    token=settings.SLACK_BOT_TOKEN,
    signing_secret=settings.SLACK_SIGNING_SECRET,
    token_verification_enabled=False,  # Not to bother dev tasks
)



"""Abstraction for dictionary templates."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


class BaseTemplate(ABC):
    """Abstract base class for dictionary templates."""

    @abstractmethod
    def render(self, *, context: dict[str, Any] | None = None) -> dict:
        """Render template with given context."""



from __future__ import annotations

from datetime import datetime
from typing import Any

from factory import lazy_attribute

from tests._factories import SlackResponseFactory


class SlackMessageResponseFactory(SlackResponseFactory):
    @lazy_attribute
    def data(self) -> dict[str, Any]:
        ts = str(datetime.now().timestamp())  # noqa: DTZ005
        return {
            "ok": True,
            "channel": "whatever-channel",
            "ts": ts,
            "message": {
                "bot_id": "<REDACTED>",
                "type": "message",
                "text": self.text,
                "user": "<REDACTED>",
                "ts": ts,
                "app_id": "<REDACTED>",
                "blocks": self.blocks,
                "team": "<REDACTED>",
                "bot_profile": {},
                "attachments": self.attachments,
            },
        }

    class Params:
        text = "Hello, World!"
        blocks = [  # noqa: RUF012
            {
                "type": "rich_text",
                "block_id": "FbU8",
                "elements": [
                    {
                        "type": "rich_text_section",
                        "elements": [
                            {
                                "type": "text",
                                "text": "Hello, World!",
                            },
                        ],
                    },
                ],
            },
        ]
        attachments = []  # type: ignore[var-annotated]  # noqa: RUF012



# noqa: D100
from __future__ import annotations

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoSlackBotConfig(AppConfig):  # noqa: D101
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_slack_tools.slack_messages"
    verbose_name = _("Slack Messages")

    def ready(self) -> None:  # pragma: no cover
        """Auto-discover Celery tasks, if Celery is installed."""
        try:
            import celery  # noqa: F401
        except ImportError:
            pass
        else:
            from . import tasks  # noqa: F401



# Generated by Django 4.2.11 on 2024-04-09 13:39
from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import migrations

if TYPE_CHECKING:
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor
    from django.db.migrations.state import StateApps


def _create_default_policy(apps: StateApps, schema_editor: BaseDatabaseSchemaEditor) -> None:
    db_alias = schema_editor.connection.alias
    model_type = apps.get_model("slack_messages", "SlackMessagingPolicy")
    model_type.objects.using(db_alias).create(code="DEFAULT", enabled=False)


def _delete_default_policy(apps: StateApps, schema_editor: BaseDatabaseSchemaEditor) -> None:
    db_alias = schema_editor.connection.alias
    model_type = apps.get_model("slack_messages", "SlackMessagingPolicy")
    model_type.objects.using(db_alias).filter(code="DEFAULT").delete()


class Migration(migrations.Migration):
    dependencies = [
        ("slack_messages", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(
            code=_create_default_policy,
            reverse_code=_delete_default_policy,
        ),
    ]



from __future__ import annotations

from typing import Any

import pytest

from django_slack_tools.slack_messages.models import SlackMention
from tests._helpers import ModelTestBase

from ._factories import SlackMentionFactory


class TestSlackMention(ModelTestBase):
    model_cls = SlackMention
    factory_cls = SlackMentionFactory

    @pytest.mark.parametrize(
        argnames=["kwargs", "expect"],
        argvalues=[
            [
                {"type": SlackMention.MentionType.SPECIAL, "name": "Here", "mention_id": "<!here>"},
                "Here (Special, <!here>)",
            ],
            [
                {"type": SlackMention.MentionType.USER, "name": "lasuillard", "mention_id": "U0000000000"},
                "lasuillard (User, U0000000000)",
            ],
            [
                {"type": SlackMention.MentionType.GROUP, "name": "Backend", "mention_id": "T0000000000"},
                "Backend (Group, T0000000000)",
            ],
        ],
        ids=["special", "user", "team"],
    )
    def test_str(self, kwargs: dict[str, Any], expect: str) -> None:
        instance = self.factory_cls.build(**kwargs)
        assert str(instance) == expect

    @pytest.mark.parametrize(
        argnames=["kwargs", "expect"],
        argvalues=[
            [
                {"type": SlackMention.MentionType.SPECIAL, "name": "Here", "mention_id": "<!here>"},
                "<!here>",
            ],
            [
                {"type": SlackMention.MentionType.USER, "name": "lasuillard", "mention_id": "U0000000000"},
                "<@U0000000000>",
            ],
            [
                {"type": SlackMention.MentionType.GROUP, "name": "Backend", "mention_id": "T0000000000"},
                "<!subteam^T0000000000>",
            ],
        ],
        ids=["special", "user", "team"],
    )
    def test_mention(self, kwargs: dict[str, Any], expect: str) -> None:
        instance = self.factory_cls.build(**kwargs)
        assert instance.mention == expect



import os
from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "django-insecure-kv)x7pa__5ls#_zi*-hgi-mi@&&v7*ebtg3vmfvlxg*a###*ug"  # noqa: S105

DEBUG = True

ALLOWED_HOSTS = ["*"]

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_celery_results",
    "django_celery_beat",
    "django_slack_tools.slack_messages",
    "todo_app",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "todo.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "todo.wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "postgres",
        "USER": "postgres",
        "PASSWORD": "password",
        "HOST": "localhost",
        "PORT": 5432,
    },
}


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "{asctime} {levelname:<9s} {name} {message}",
            "style": "{",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "simple",
        },
    },
    "root": {
        "handlers": ["console"],
        "level": "DEBUG",
        "propagate": True,
    },
}

SLACK_BOT_TOKEN = os.environ.get("SLACK_BOT_TOKEN", default="i-am-a-cookie")
SLACK_SIGNING_SECRET = os.environ.get("SLACK_SIGNING_SECRET", default="stupid-potato")

DJANGO_SLACK_TOOLS = {
    "SLACK_APP": "todo.slack_app.app",
    "BACKEND": {
        "NAME": "django_slack_tools.slack_messages.backends.SlackBackend",
        "OPTIONS": {
            "slack_app": "todo.slack_app.app",
        },
    },
}



import pytest
from django.core.exceptions import ValidationError

from django_slack_tools.utils.slack import body_validator, header_validator


def test_header_validator() -> None:
    header_validator({})
    with pytest.raises(ValidationError):
        header_validator(
            {
                "reply_broadcast": "Give me an error",
            },
        )


def test_body_validator() -> None:
    body_validator(
        {
            "text": "Hello, World!",
        },
    )
    with pytest.raises(ValidationError):
        body_validator({})



# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from django.contrib import admin, messages
from django.contrib.admin.filters import ChoicesFieldListFilter, DateFieldListFilter
from django.utils.translation import gettext_lazy as _

from django_slack_tools.app_settings import app_settings
from django_slack_tools.slack_messages.models import SlackMention

if TYPE_CHECKING:
    from django.db.models.query import QuerySet
    from django.http import HttpRequest


@admin.register(SlackMention)
class SlackMentionAdmin(admin.ModelAdmin):
    """Admin for recipients."""

    readonly_fields = ("id", "_get_mention_str", "created", "last_modified")

    @admin.display(description=_("Mention string"))
    def _get_mention_str(self, instance: SlackMention) -> str:
        return instance.mention

    # Actions
    actions = ("_update_mentions",)

    @admin.action(description=_("Update type and name of mentions"))
    def _update_mentions(self, request: HttpRequest, queryset: QuerySet[SlackMention]) -> None:
        """Admin action to update selected mentions' type and name using Slack workspace information."""
        items = _get_mentionable_items()

        # Iterate over all make temporary updates
        changes: list[SlackMention] = []
        failures: list[SlackMention] = []
        for mention in queryset:
            match = items.get(mention.mention_id)
            if match is None:
                failures.append(mention)
                continue

            mention.type = match["type"]
            mention.name = match["name"]
            changes.append(mention)

        # Make changes in bulk
        SlackMention.objects.bulk_update(changes, fields=("type", "name"))
        n_success = len(changes)

        # Report back to admin
        if failures:
            messages.warning(
                request,
                _(
                    "Updated {n_success} mentions successfully"
                    " and there were {n_fail} mentions failed to update because no matching data.",
                ).format(n_success=n_success, n_fail=len(failures)),
            )
        else:
            messages.info(request, _("Updated {n} mentions.").format(n=n_success))

    # Changelist
    # ------------------------------------------------------------------------
    date_hierarchy = "last_modified"
    search_fields = ("name", "mention_id")
    list_display = ("id", "name", "type", "mention_id", "created", "last_modified")
    list_display_links = ("id", "name", "mention_id")
    list_filter = (
        ("type", ChoicesFieldListFilter),
        ("created", DateFieldListFilter),
        ("last_modified", DateFieldListFilter),
    )

    # Change
    # ------------------------------------------------------------------------
    fieldsets = (
        (
            None,
            {
                "fields": ("name", "type", "mention_id", "_get_mention_str"),
            },
        ),
        (
            _("Miscellaneous"),
            {
                "fields": ("id", "created", "last_modified"),
                "classes": ("collapse",),
            },
        ),
    )


def _get_mentionable_items() -> dict[str, _Mentionable]:
    """Returns mapping of ID to mention name for members and usergroups."""
    slack_app = app_settings.slack_app

    # List of mentionable
    items: dict[str, _Mentionable] = {}

    # Fetch members from Slack
    # TODO(lasuillard): Need pagination in future
    response = slack_app.client.users_list()
    members: list[dict] = response.get("members", default=[])

    # Members mapping
    items.update(
        {
            member["id"]: {
                "type": SlackMention.MentionType.USER,
                "name": member["profile"]["display_name"] or member["profile"]["real_name"],
            }
            for member in members
        },
    )

    # Fetch usergroups from Slack
    response = slack_app.client.usergroups_list()
    usergroups: list[dict] = response.get("usergroups", default=[])

    # Usergroups mapping
    items.update(
        {
            usergroup["id"]: {
                "type": SlackMention.MentionType.GROUP,
                "name": usergroup["name"],
            }
            for usergroup in usergroups
        },
    )

    return items


class _Mentionable(TypedDict):
    type: SlackMention.MentionType
    name: str



"""Dummy backend doing nothing."""

from __future__ import annotations

from logging import getLogger
from typing import Any

from slack_sdk.web import SlackResponse

from django_slack_tools.slack_messages.models import SlackMessage

from .base import BaseBackend

logger = getLogger(__name__)


class DummyBackend(BaseBackend):
    """An dummy backend that does nothing with message."""

    def prepare_message(self, *args: Any, **kwargs: Any) -> SlackMessage:  # noqa: D102, ARG002
        return SlackMessage(header={}, body={})

    def _send_message(self, *args: Any, **kwargs: Any) -> SlackResponse:  # noqa: ARG002
        return SlackResponse(
            client=None,
            http_verb="POST",
            api_url="https://www.slack.com/api/chat.postMessage",
            req_args={},
            data={"ok": True},
            headers={},
            status_code=200,
        )

    def _get_permalink(self, *, message: SlackMessage, raise_exception: bool) -> str:  # noqa: ARG002
        return ""

    def _record_request(self, *args: Any, **kwargs: Any) -> Any: ...

    def _record_response(self, *args: Any, **kwargs: Any) -> Any: ...



import pytest

from django_slack_tools.utils.slack.message import MessageBody, MessageHeader


class TestMessageHeader:
    def test_instance_creation(self) -> None:
        assert MessageHeader()

    def test_from_any(self) -> None:
        assert MessageHeader.from_any(None) == MessageHeader()
        assert MessageHeader.from_any(
            {"mrkdwn": "some-markdown"},
        ) == MessageHeader(mrkdwn="some-markdown")
        with pytest.raises(TypeError, match="Unsupported type <class 'int'>"):
            MessageHeader.from_any(-1)  # type: ignore[arg-type]


class TestMessageBody:
    def test_instance_creation(self) -> None:
        assert MessageBody(text="some-text")

    def test_from_any(self) -> None:
        assert MessageBody.from_any({"text": "some-text"}) == MessageBody(text="some-text")
        assert MessageBody.from_any("some-text") == MessageBody(text="some-text")
        with pytest.raises(TypeError, match="Unsupported type <class 'int'>"):
            MessageBody.from_any(-1)  # type: ignore[arg-type]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from django_slack_tools.slack_messages.admin import SlackMessagingPolicyAdmin
from django_slack_tools.slack_messages.models import SlackMessagingPolicy
from tests._helpers import ModelAdminTestBase
from tests.slack_messages.models._factories import SlackMessagingPolicyFactory

if TYPE_CHECKING:
    from django.test.client import Client


class TestSlackMessagingPolicyAdmin(ModelAdminTestBase):
    admin_cls = SlackMessagingPolicyAdmin
    model_cls = SlackMessagingPolicy
    factory_cls = SlackMessagingPolicyFactory

    @pytest.mark.parametrize(
        argnames="kwargs",
        argvalues=[
            {"template": {"blocks": [{"type": "section", "text": {"type": "plain_text", "text": "Hello"}}]}},
            {"template": {"attachments": [{"type": "section", "text": {"type": "plain_text", "text": "Hello"}}]}},
            {
                "template": {
                    "blocks": [{"type": "section", "text": {"type": "plain_text", "text": "Hello"}}],
                    "attachments": [{"type": "section", "text": {"type": "plain_text", "text": "Hello"}}],
                },
            },
        ],
        ids=["blocks only", "attachments only", "both blocks and attachments"],
    )
    def test_change(self, admin_client: Client, kwargs: dict) -> None:
        obj = self.factory_cls.create(**kwargs)
        url = self._reverse("change", kwargs={"object_id": obj.id})

        # Test visit
        response = admin_client.get(url)
        assert response.status_code == 200



# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING

import django.contrib.messages
from django.contrib import admin
from django.contrib.admin.filters import DateFieldListFilter
from django.db import models
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from django_slack_tools.app_settings import app_settings
from django_slack_tools.slack_messages.models import SlackMessage
from django_slack_tools.utils.widgets import JSONWidget

if TYPE_CHECKING:
    from django.db.models.query import QuerySet
    from django.http import HttpRequest
    from django_stubs_ext import StrOrPromise


@admin.register(SlackMessage)
class SlackMessageAdmin(admin.ModelAdmin):
    """Admin for messages."""

    readonly_fields = ("id", "_link_to_permalink", "created", "last_modified")

    @admin.display(description=_("Permalink"))
    def _link_to_permalink(self, instance: SlackMessage) -> StrOrPromise:
        url = instance.permalink
        if not url:
            return _("N/A")

        return format_html("<a href='{url}'>{title}</a>", url=url, title=_("Permalink"))

    # Actions
    actions = ("_clone_messages", "_send_messages")
    actions_on_bottom = True  # Likely to be there are lots of messages

    @admin.action(description=_("Clone messages"))
    def _clone_messages(self, request: HttpRequest, queryset: QuerySet[SlackMessage]) -> None:
        messages = list(queryset)
        for message in queryset:
            message.pk = None
            message.ts = None
            message.parent_ts = ""
            message.ok = None
            message.request = None
            message.response = None

        created = SlackMessage.objects.bulk_create(messages)
        n = len(created)
        django.contrib.messages.info(request, _("Cloned {n} messages.").format(n=n))

    @admin.action(description=_("Send messages"))
    def _send_messages(self, request: HttpRequest, queryset: QuerySet[SlackMessage]) -> None:
        backend = app_settings.backend

        # Send selected messages
        n_success = 0
        n_failure = 0
        for message in queryset:
            backend.send_message(message, raise_exception=False, get_permalink=False)
            if message.ok:
                n_success += 1
            else:
                n_failure += 1

        # Notify result to user
        n = n_success + n_failure
        if n_failure:
            django.contrib.messages.info(
                request,
                _("Tried to send {n} messages, {n_success} succeeded and {n_failure} failed.").format(
                    n=n,
                    n_success=n_success,
                    n_failure=n_failure,
                ),
            )
        else:
            django.contrib.messages.info(request, _("Sent {n} messages.").format(n=n))

    # Changelist
    # ------------------------------------------------------------------------
    date_hierarchy = "created"
    search_fields = ("id", "ts", "parent_ts", "policy__code", "channel")
    list_display = (
        "id",
        "ts",
        "ok",
        "policy",
        "channel",
        "parent_ts",
        "_link_to_permalink",
        "created",
        "last_modified",
    )
    list_display_links = ("id", "ts")
    list_select_related = ("policy",)
    list_filter = (
        "policy",
        "ok",
        ("created", DateFieldListFilter),
        ("last_modified", DateFieldListFilter),
    )

    # Change
    # ------------------------------------------------------------------------
    formfield_overrides = {  # noqa: RUF012
        models.JSONField: {"widget": JSONWidget},
    }
    fieldsets = (
        (
            None,
            {
                "fields": ("policy", "channel", "ok", "ts", "parent_ts", "_link_to_permalink", "header", "body"),
            },
        ),
        (
            _("Miscellaneous"),
            {
                "fields": ("id", "request", "response", "created", "last_modified"),
                "classes": ("collapse",),
            },
        ),
    )
    autocomplete_fields = ("policy",)



from datetime import datetime
from typing import Any, Sequence

import faker
from django.utils import timezone
from factory import Faker, LazyAttribute, post_generation
from factory.django import DjangoModelFactory

from django_slack_tools.slack_messages.models import (
    SlackMention,
    SlackMessage,
    SlackMessageRecipient,
    SlackMessagingPolicy,
)

_fake = faker.Faker()


class SlackMentionFactory(DjangoModelFactory):
    class Meta:
        model = SlackMention

    type = SlackMention.MentionType.UNKNOWN
    name = Faker("name")
    mention_id = Faker("pystr", max_chars=12)


class SlackMessageRecipientFactory(DjangoModelFactory):
    class Meta:
        model = SlackMessageRecipient

    alias = Faker("name")
    channel = Faker("pystr", max_chars=12)
    channel_name = LazyAttribute(lambda _: f"#{_fake.pystr()}")

    @post_generation
    def mentions(
        self: SlackMessageRecipient,
        create: bool,  # noqa: FBT001
        extracted: Sequence[SlackMention],
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        if not create or not extracted:
            return

        self.mentions.add(*extracted)


class SlackMessageFactory(DjangoModelFactory):
    class Meta:
        model = SlackMessage

    policy = None
    channel = LazyAttribute(lambda _: f"#{_fake.pystr()}")
    header: dict = {}  # noqa: RUF012
    body = LazyAttribute(lambda _: {"text": _fake.paragraph()})
    ok = None
    ts = LazyAttribute(lambda _: str(timezone.now().timestamp()))
    parent_ts = ""
    request = None
    response = None

    # Hook to set the created(`auto_now_add` set) field after creation
    @post_generation
    def created(self: SlackMessage, create: bool, extracted: datetime, **kwargs: Any) -> None:  # noqa: ARG002, FBT001
        if not create or not extracted:
            return

        self.created = extracted
        self.save()


class SlackMessagingPolicyFactory(DjangoModelFactory):
    class Meta:
        model = SlackMessagingPolicy

    code = Faker("pystr")
    enabled = True

    @post_generation
    def recipients(
        self: SlackMessagingPolicy,
        create: bool,  # noqa: FBT001
        extracted: Sequence[SlackMessageRecipient],
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        if not create or not extracted:
            return

        self.recipients.add(*extracted)

    template_type = SlackMessagingPolicy.TemplateType.DICT
    template = {  # noqa: RUF012
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "Hello, World!: {mentions}",
                },
            },
        ],
        "attachments": [
            {
                "color": "#f2c744",
                "blocks": [{"type": "section", "text": {"type": "mrkdwn", "text": "Hello, World!: {mentions}"}}],
            },
        ],
    }
    header_defaults: dict = {}  # noqa: RUF012



from django_slack_tools.slack_messages.message import slack_message_via_policy

from .models import Todo


def create_todo(title: str, description: str) -> Todo:
    """Create a new To Do."""
    todo = Todo.objects.create(title=title, description=description)  # type: ignore[attr-defined]
    slack_message_via_policy("TODO-CREATED", lazy=True, context={"title": todo.title})
    return todo  # type:ignore[no-any-return]


def mark_todo_as_completed(todo_id: int) -> None:
    """Mark a To Do as completed."""
    todo = Todo.objects.get(id=todo_id)  # type: ignore[attr-defined]
    todo.completed = True
    todo.save()
    slack_message_via_policy("TODO-COMPLETED", lazy=True, context={"title": todo.title})



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

import pytest
from django.db import models

from django_slack_tools.slack_messages.admin import SlackMessageAdmin
from django_slack_tools.slack_messages.models import SlackMessage
from tests._factories import SlackApiErrorFactory, SlackResponseFactory
from tests._helpers import ModelAdminTestBase
from tests.slack_messages.models._factories import SlackMessageFactory

if TYPE_CHECKING:
    from unittest.mock import Mock

    from django.test import Client
    from django.test.client import _MonkeyPatchedWSGIResponse


class TestSlackMessageAdmin(ModelAdminTestBase):
    admin_cls = SlackMessageAdmin
    model_cls = SlackMessage
    factory_cls = SlackMessageFactory

    pytestmark = pytest.mark.django_db()

    def _clone_messages(
        self,
        *,
        client: Client,
        ids: Iterable[SlackMessage],
    ) -> _MonkeyPatchedWSGIResponse:
        return client.post(
            self._reverse("changelist"),
            {
                "action": "_clone_messages",
                "_selected_action": ids,
            },
        )

    def _send_messages(
        self,
        *,
        client: Client,
        ids: Iterable[SlackMessage],
    ) -> _MonkeyPatchedWSGIResponse:
        return client.post(
            self._reverse("changelist"),
            {
                "action": "_send_messages",
                "_selected_action": ids,
            },
        )

    def test_changelist(self, admin_client: Client) -> None:
        # Test permalink field
        self.factory_cls.create(permalink="https://example.com")
        return super().test_changelist(admin_client)

    def test_clone_messages(self, admin_client: Client) -> None:
        # These messages should clone
        messages_to_clone = self.factory_cls.create_batch(size=3)
        ids = [m.id for m in messages_to_clone]

        # Perform action
        response = self._clone_messages(client=admin_client, ids=ids)
        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == ["Cloned 3 messages."]

        # Check messages are cloned
        assert SlackMessage.objects.count() == 6

    def test_send_messages(self, admin_client: Client, mock_slack_client: Mock) -> None:
        # These messages should send
        messages_to_send = self.factory_cls.create_batch(size=3)
        ids = [m.id for m in messages_to_send]

        # Perform action
        mock_slack_client.chat_postMessage.return_value = SlackResponseFactory()
        response = self._send_messages(client=admin_client, ids=ids)

        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == ["Sent 3 messages."]

        # Check messages are sent
        assert SlackMessage.objects.count() == 3
        assert SlackMessage.objects.filter(ok=True).count() == 3

    def test_send_messages_partial_failure(self, admin_client: Client, mock_slack_client: Mock) -> None:
        # These messages should send
        messages_to_send = self.factory_cls.create_batch(size=3)
        ids = [m.id for m in messages_to_send]

        # Perform action
        mock_slack_client.chat_postMessage.side_effect = [
            SlackResponseFactory(),
            SlackApiErrorFactory(),
            SlackResponseFactory(),
        ]
        response = self._send_messages(client=admin_client, ids=ids)

        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == ["Tried to send 3 messages, 2 succeeded and 1 failed."]

        # Check messages are sent
        messages_from_db = SlackMessage.objects.values("ok").annotate(count=models.Count("ok")).order_by("ok")
        assert list(messages_from_db) == [
            {"ok": False, "count": 1},
            {"ok": True, "count": 2},
        ]



# noqa: D100
from __future__ import annotations

import logging
import re
import xml.etree.ElementTree as ET
from textwrap import dedent
from typing import TYPE_CHECKING, overload

import xmltodict
from django.template import engines

from .base import BaseTemplate

if TYPE_CHECKING:
    from typing import Any

    from django.template.backends.base import BaseEngine

logger = logging.getLogger(__name__)


class DjangoTemplate(BaseTemplate):
    """Template utilizing Django built-in template engine."""

    @overload
    def __init__(self, *, file: str, engine: BaseEngine | None = None) -> None: ...  # pragma: no cover

    @overload
    def __init__(self, *, inline: str, engine: BaseEngine | None = None) -> None: ...  # pragma: no cover

    def __init__(
        self,
        *,
        file: str | None = None,
        inline: str | None = None,
        engine: BaseEngine | None = None,
    ) -> None:
        """Initialize template.

        Args:
            file: Path to file with template.
            inline: XML inline template.
            engine: Template engine to use. Defaults to Django engine.

        Raises:
            TypeError: Some of the arguments are missing or multiple are provided.
            ValueError: Unsupported value provided.
        """
        engine = engines["django"] if engine is None else engine

        if len([value for value in (file, inline) if value is not None]) != 1:
            msg = "Exactly one of 'file' or 'inline' must be provided."
            raise TypeError(msg)

        if file:
            template = engine.get_template(file)
        elif inline:
            template = engine.from_string(inline)
        else:  # pragma: no cover
            msg = "Unreachable code"
            raise NotImplementedError(msg)

        self.template = template

    def render(self, *, context: dict[str, Any] | None = None) -> dict:  # noqa: D102
        context = {} if context is None else context

        logger.debug("Rendering template with context: %r", context)
        rendered = self.template.render(context=context)
        return _xml_to_dict(rendered)


def _xml_to_dict(xml: str) -> dict:
    """Parse XML string to Python dictionary.

    Following transformations are applied by default:

    - Normalize text nodes: remove single newlines and dedent text, etc.
    - Rename tags for syntactic comfort: block -> blocks, element -> elements

    Please check the tests for more detailed examples.

    Args:
        xml: XML string.

    Returns:
        Parsed dictionary. Be aware, the returned value will be the child of
        top-level node (e.g. <root>...</root>), regardless of its key name.
    """
    xml = _preprocess_xml(xml)
    obj = xmltodict.parse(
        xml,
        attr_prefix="",
        cdata_key="text",
        force_list=("blocks", "elements"),
        postprocessor=_xml_postprocessor,
    )
    return dict(next(iter(obj.values())))


def _preprocess_xml(xml: str) -> str:
    """Normalize XML text nodes."""
    root = ET.fromstring(xml)  # noqa: S314; TODO(lasuillard): Naive belief that XML is safe
    for node in root.iter():
        node.tag = _rename_tag(node.tag)

        if node.tag == "text" and node.text:
            text = dedent(node.text)
            text = _remove_single_newline(text)
            logger.debug("Normalized text node: %r -> %r", node.text, text)
            node.text = text

    return ET.tostring(root, encoding="unicode")


def _rename_tag(tag: str) -> str:
    """Rename tags."""
    if tag == "block":
        return "blocks"

    if tag == "element":
        return "elements"

    return tag


def _remove_single_newline(text: str) -> str:
    """Remove a single newline from repeated newlines. If the are just one newline, replace it with space."""
    return re.sub(r"([\n]+)", lambda m: "\n" * (m.group(1).count("\n") - 1) or " ", text)


def _xml_postprocessor(path: Any, key: str, value: Any) -> tuple[str, Any]:  # noqa: ARG001
    if value == "true":
        return key, True

    if value == "false":
        return key, False

    return key, value



"""Logging backend that do logging otherwise nothing."""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any

from .dummy import DummyBackend

if TYPE_CHECKING:
    from slack_sdk.web import SlackResponse


logger = getLogger(__name__)


class LoggingBackend(DummyBackend):
    """Backend that log the message rather than sending it."""

    def _send_message(self, *args: Any, **kwargs: Any) -> SlackResponse:
        logger.debug("Sending an message with following args=%r, kwargs=%r", args, kwargs)
        return super()._send_message(*args, **kwargs)

    def _record_request(self, *args: Any, **kwargs: Any) -> Any:
        logger.debug("Recording request with args=%r, kwargs=%r", args, kwargs)

    def _record_response(self, *args: Any, **kwargs: Any) -> Any:
        logger.debug("Recording response with args=%r, kwargs=%r", args, kwargs)



from .django import body_validator, header_validator
from .message import MessageBody, MessageHeader
from .misc import get_block_kit_builder_url

__all__ = (
    "body_validator",
    "header_validator",
    "MessageBody",
    "MessageHeader",
    "get_block_kit_builder_url",
)



# Generated by Django 3.2.23 on 2024-01-25 12:57

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Todo",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="Title of todo.",
                        max_length=100,
                        verbose_name="Title",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Description of todo.",
                        verbose_name="Description",
                    ),
                ),
                (
                    "completed",
                    models.BooleanField(
                        default=False,
                        help_text="Is todo completed?",
                        verbose_name="Completed",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="When instance created.",
                        verbose_name="Created",
                    ),
                ),
                (
                    "last_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When instance modified recently.",
                        verbose_name="Last Modified",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.2.16 on 2024-10-19 05:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("slack_messages", "0003_default_recipient"),
    ]

    operations = [
        migrations.AddField(
            model_name="slackmessagingpolicy",
            name="template_type",
            field=models.CharField(
                choices=[("D", "Dictionary"), ("DJ", "Django"), ("DI", "Django Inline"), ("?", "Unknown")],
                default="D",
                help_text="Type of message template.",
                max_length=2,
                verbose_name="Template type",
            ),
        ),
    ]



"""Some stuffs for Django such as field validators."""

from __future__ import annotations

from django.core.exceptions import ValidationError

from .message import MessageBody, MessageHeader


def header_validator(d: dict) -> None:
    """Validate given value is valid message header."""
    try:
        MessageHeader.from_any(d)
    except Exception as exc:
        err = _convert_errors(exc)
        raise err from exc


def body_validator(d: dict) -> None:
    """Validate given value is valid message body."""
    try:
        MessageBody.from_any(d)
    except Exception as exc:
        err = _convert_errors(exc)
        raise err from exc


def _convert_errors(exc: Exception) -> ValidationError:
    return ValidationError(str(exc))



"""Handy APIs for sending Slack messages."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django_slack_tools.app_settings import app_settings
from django_slack_tools.slack_messages.models.message_recipient import SlackMessageRecipient
from django_slack_tools.utils.slack import MessageBody, MessageHeader

from .models import SlackMessagingPolicy

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from django_slack_tools.slack_messages.backends.base import BaseBackend

    from .models import SlackMessage


def slack_message(  # noqa: PLR0913
    body: str | MessageBody | dict[str, Any],
    *,
    channel: str,
    header: MessageHeader | dict[str, Any] | None = None,
    raise_exception: bool = False,
    get_permalink: bool = False,
    backend: BaseBackend = app_settings.backend,
) -> SlackMessage:
    """Send a simple text message.

    Args:
        body: Message content, simple message or full request body.
        channel: Channel to send message.
        header: Slack message control header.
        raise_exception: Whether to re-raise caught exception while sending messages.
        get_permalink: Try to get the message permalink via extraneous Slack API calls.
        backend: Messaging backend. If not set, use `app_settings.backend`.

    Returns:
        Sent message instance or `None`.
    """
    body = MessageBody.from_any(body)
    header = MessageHeader.from_any(header)
    message = backend.prepare_message(channel=channel, header=header, body=body)

    return backend.send_message(message, raise_exception=raise_exception, get_permalink=get_permalink)


def slack_message_via_policy(  # noqa: PLR0913
    policy: str | SlackMessagingPolicy = app_settings.default_policy_code,
    *,
    header: MessageHeader | dict[str, Any] | None = None,
    raise_exception: bool = False,
    lazy: bool = False,
    get_permalink: bool = False,
    context: dict[str, Any] | None = None,
    backend: BaseBackend = app_settings.backend,
) -> int:
    """Send a simple text message.

    Some default context variables are populated and available for use in templates.
    See corresponding backend implementation for available context variables.

    Args:
        policy: Messaging policy code or policy instance. Defaults to app's default policy.
        header: Slack message control header.
        raise_exception: Whether to re-raise caught exception while sending messages.
        lazy: Decide whether try create policy with disabled, if not exists.
        get_permalink: Try to get the message permalink via extraneous Slack API calls.
        context: Dictionary to pass to template for rendering.
        backend: Messaging backend. If not set, use `app_settings.backend`.

    Returns:
        Count of messages sent successfully.

    Raises:
        SlackMessagingPolicy.DoesNotExist: Policy for given code does not exists.
    """
    if isinstance(policy, str):
        if lazy:
            policy, created = SlackMessagingPolicy.objects.get_or_create(
                code=policy,
                defaults={
                    "enabled": app_settings.lazy_policy_enabled,
                    "template": app_settings.default_template,
                },
            )
            if created:
                # Add default recipient for created policy
                recipient = SlackMessageRecipient.objects.get(alias=app_settings.default_recipient)
                policy.recipients.add(recipient)
                logger.warning("Policy for code %r created because `lazy` is set.", policy)
        else:
            policy = SlackMessagingPolicy.objects.get(code=policy)

    header = MessageHeader.from_any(header)
    context = context or {}

    messages = backend.prepare_messages_from_policy(policy, header=header, context=context)
    if not policy.enabled:
        logger.warning(
            "Created %d messages but not sending because policy %s is not enabled.",
            len(messages),
            policy.code,
        )
        return 0

    return backend.send_messages(*messages, raise_exception=raise_exception, get_permalink=get_permalink)



from .mention import SlackMentionAdmin
from .message import SlackMessageAdmin
from .message_recipient import SlackMessageRecipientAdmin
from .messaging_policy import SlackMessagingPolicyAdmin

__all__ = ("SlackMentionAdmin", "SlackMessageAdmin", "SlackMessageRecipientAdmin", "SlackMessagingPolicyAdmin")



from __future__ import annotations

import pytest

from django_slack_tools.slack_messages.backends import DummyBackend
from django_slack_tools.utils.slack.message import MessageBody, MessageHeader


class TestDummyBackend:
    pytestmark = pytest.mark.django_db()

    def test_backend(self) -> None:
        backend = DummyBackend()
        message = backend.prepare_message(
            channel="dummy-dummy",
            header=MessageHeader(),
            body=MessageBody(text="Get some sleep"),
        )
        backend.send_message(message=message, raise_exception=True, get_permalink=True)



"""Slack messaging backends."""

from __future__ import annotations

import dataclasses
import traceback
from abc import ABC, abstractmethod
from logging import getLogger
from typing import TYPE_CHECKING, Any, cast

from slack_sdk.errors import SlackApiError

from django_slack_tools.slack_messages.models import SlackMessage, SlackMessagingPolicy
from django_slack_tools.utils.slack import MessageBody
from django_slack_tools.utils.template import DictTemplate, DjangoTemplate

if TYPE_CHECKING:
    from slack_sdk.web import SlackResponse

    from django_slack_tools.slack_messages.models.mention import SlackMention
    from django_slack_tools.slack_messages.models.message_recipient import SlackMessageRecipient
    from django_slack_tools.utils.slack import MessageHeader
    from django_slack_tools.utils.template import BaseTemplate

logger = getLogger(__name__)

RESERVED_CONTEXT_KWARGS = frozenset({"policy", "mentions", "mentions_as_str"})
"""Set of reserved context keys automatically created."""


class BaseBackend(ABC):
    """Abstract base class for messaging backends."""

    def prepare_message(
        self,
        *,
        policy: SlackMessagingPolicy | None = None,
        channel: str,
        header: MessageHeader,
        body: MessageBody,
    ) -> SlackMessage:
        """Prepare message.

        Args:
            policy: Related policy instance.
            channel: Channel to send message.
            header: Slack message control header.
            body: Slack message body.

        Returns:
            Prepared message.
        """
        _header: dict = policy.header_defaults if policy else {}
        _header.update(dataclasses.asdict(header))

        _body = dataclasses.asdict(body)

        return SlackMessage(policy=policy, channel=channel, header=_header, body=_body)

    def prepare_messages_from_policy(
        self,
        policy: SlackMessagingPolicy,
        *,
        header: MessageHeader,
        context: dict[str, Any],
    ) -> list[SlackMessage]:
        """Prepare messages from policy.

        Args:
            policy: Policy to create messages from.
            header: Common message header.
            context: Message context.

        Returns:
            Prepared messages.
        """
        overridden_reserved = RESERVED_CONTEXT_KWARGS & set(context.keys())
        if overridden_reserved:
            logger.warning(
                "Template keyword argument(s) %s reserved for passing mentions, but already exists."
                " User provided value will override it.",
                ", ".join(f"`{s}`" for s in overridden_reserved),
            )

        messages: list[SlackMessage] = []
        for recipient in policy.recipients.all():
            logger.debug("Sending message to recipient %s", recipient)

            # Initialize template instance
            template = self._get_template_instance_from_policy(policy)

            # Prepare rendering arguments
            render_context = self._get_default_context(policy=policy, recipient=recipient)
            render_context.update(context)
            logger.debug("Context prepared as: %r", render_context)

            # Render template and parse as body
            rendered = template.render(context=render_context)
            body = MessageBody.from_any(rendered)

            # Create message instance
            message = self.prepare_message(policy=policy, channel=recipient.channel, header=header, body=body)
            messages.append(message)

        return SlackMessage.objects.bulk_create(messages)

    def _get_template_instance_from_policy(self, policy: SlackMessagingPolicy) -> BaseTemplate:
        """Get template instance."""
        if policy.template_type == SlackMessagingPolicy.TemplateType.DICT:
            return DictTemplate(policy.template)

        if policy.template_type == SlackMessagingPolicy.TemplateType.DJANGO:
            return DjangoTemplate(file=policy.template)

        if policy.template_type == SlackMessagingPolicy.TemplateType.DJANGO_INLINE:
            return DjangoTemplate(inline=policy.template)

        msg = f"Unsupported template type: {policy.template_type!r}"
        raise ValueError(msg)

    def _get_default_context(self, *, policy: SlackMessagingPolicy, recipient: SlackMessageRecipient) -> dict[str, Any]:
        """Get default context for rendering.

        Following default context keys are created:

        - `policy`: Policy code.
        - `mentions`: List of mentions.
        - `mentions_as_str`: Comma-separated joined string of mentions.
        """
        mentions: list[SlackMention] = list(recipient.mentions.all())
        mentions_as_str = ", ".join(mention.mention for mention in mentions)

        return {
            "policy": policy.code,
            "mentions": mentions,
            "mentions_as_str": mentions_as_str,
        }

    def send_messages(self, *messages: SlackMessage, raise_exception: bool, get_permalink: bool) -> int:
        """Shortcut to send multiple messages.

        Args:
            messages: Messages to send.
            raise_exception: Whether to propagate exceptions.
            get_permalink: Try to get the message permalink via additional Slack API call.

        Returns:
            Count of messages sent successfully.
        """
        num_sent = 0
        for message in messages:
            sent = self.send_message(message=message, raise_exception=raise_exception, get_permalink=get_permalink)
            num_sent += 1 if sent.ok else 0

        return num_sent

    def send_message(
        self,
        message: SlackMessage,
        *,
        raise_exception: bool,
        get_permalink: bool,
    ) -> SlackMessage:
        """Send message.

        Args:
            message: Prepared message.
            raise_exception: Whether to propagate exceptions.
            get_permalink: Try to get the message permalink via additional Slack API call.

        Returns:
            Message sent to Slack.
        """
        try:
            response: SlackResponse
            try:
                response = self._send_message(message)
            except SlackApiError as err:
                if raise_exception:
                    raise

                logger.warning(
                    "Error occurred while sending message but suppressed because `raise_exception` set.",
                    exc_info=err,
                )
                response = err.response

            message.ok = ok = cast(bool, response.get("ok"))
            if ok:
                # Get message TS if OK
                message.ts = cast(str, response.get("ts"))

                # Store thread TS if possible
                data: dict[str, Any] = response.get("message", {})
                message.parent_ts = data.get("thread_ts", "")

                # Get message permalink
                if get_permalink:
                    message.permalink = self._get_permalink(message=message, raise_exception=raise_exception)

            message.request = self._record_request(response)
            message.response = self._record_response(response)
        except:  # noqa: E722
            if raise_exception:
                raise

            logger.exception("Error occurred while sending message but suppressed because `raise_exception` set.")
            message.exception = traceback.format_exc()
        finally:
            message.save()

        message.refresh_from_db()
        return message

    @abstractmethod
    def _send_message(self, message: SlackMessage) -> SlackResponse:
        """Internal implementation of actual 'send message' behavior."""

    @abstractmethod
    def _get_permalink(self, *, message: SlackMessage, raise_exception: bool) -> str:
        """Get a permalink for given message identifier."""

    @abstractmethod
    def _record_request(self, response: SlackResponse) -> Any:
        """Extract request data to be recorded. Should return JSON-serializable object."""

    @abstractmethod
    def _record_response(self, response: SlackResponse) -> Any:
        """Extract response data to be recorded. Should return JSON-serializable object."""



from django_slack_tools.utils.template import DictTemplate


class TestDictTemplate:
    def test_render(self) -> None:
        template = DictTemplate(
            {
                "blocks": [
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": "{greet}",
                        },
                    },
                ],
                "unknown": False,
            },
        )
        result = template.render(context={"greet": "Hello, World!"})
        assert result == {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Hello, World!",
                    },
                },
            ],
            "unknown": False,
        }



from django.db import models
from django.utils.translation import gettext_lazy as _


class Todo(models.Model):
    title = models.CharField(
        verbose_name=_("Title"),
        help_text=_("Title of todo."),
        max_length=100,
    )
    description = models.TextField(
        verbose_name=_("Description"),
        help_text=_("Description of todo."),
        blank=True,
        default="",
    )
    completed = models.BooleanField(
        verbose_name=_("Completed"),
        help_text=_("Is todo completed?"),
        default=False,
    )
    created = models.DateTimeField(
        verbose_name=_("Created"),
        help_text=_("When instance created."),
        auto_now_add=True,
    )
    last_modified = models.DateTimeField(
        verbose_name=_("Last Modified"),
        help_text=_("When instance modified recently."),
        auto_now=True,
    )

    def __str__(self) -> str:
        return self.title



"""Celery utils."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from celery import shared_task
from celery.utils.log import get_task_logger
from django.utils import timezone

from django_slack_tools.slack_messages import message
from django_slack_tools.slack_messages.models import SlackMessage

if TYPE_CHECKING:
    from typing import Any

logger = get_task_logger(__name__)


@shared_task
def slack_message(
    body: str | dict[str, Any],
    *,
    channel: str,
    header: dict[str, Any] | None = None,
    raise_exception: bool = False,
    get_permalink: bool = False,
) -> int | None:
    """Celery task wrapper for `message.slack_message`.

    Args:
        body: Message content, simple message or full request body.
        channel: Channel to send message.
        header: Slack message control header.
        raise_exception: Whether to re-raise caught exception while sending messages.
        get_permalink: Try to get the message permalink via extraneous Slack API calls.

    Returns:
        ID of sent message.
    """
    sent_msg = message.slack_message(
        body,
        channel=channel,
        header=header,
        raise_exception=raise_exception,
        get_permalink=get_permalink,
    )
    return sent_msg.id


@shared_task
def slack_message_via_policy(  # noqa: PLR0913
    policy: str,
    *,
    header: dict[str, Any] | None = None,
    raise_exception: bool = False,
    lazy: bool = False,
    get_permalink: bool = False,
    context: dict[str, Any] | None = None,
) -> int:
    """Celery task wrapper for `message.slack_message_via_policy`.

    Args:
        policy: Messaging policy code.
        header: Slack message control header.
        raise_exception: Whether to re-raise caught exception while sending messages.
        lazy: Decide whether try to create policy with disabled, if not exists.
        get_permalink: Try to get the message permalink via extraneous Slack API calls.
        context: Context variables for message rendering.

    Returns:
        Number of sent messages.
    """
    return message.slack_message_via_policy(
        policy,
        header=header,
        raise_exception=raise_exception,
        lazy=lazy,
        get_permalink=get_permalink,
        context=context,
    )


@shared_task
def cleanup_old_messages(
    *,
    base_ts: str | None = None,
    threshold_seconds: int | None = 7 * 24 * 60 * 60,  # 7 days
) -> int:
    """Delete old messages created before given threshold.

    Args:
        threshold_seconds: Threshold seconds. Defaults to 7 days.
        base_ts: Base timestamp to calculate the threshold, in ISO format. If falsy, current timestamp will be used.

    Returns:
        Number of deleted messages.
    """
    dt = datetime.fromisoformat(base_ts) if base_ts else timezone.localtime()

    if threshold_seconds is None:
        logger.warning("Threshold seconds not provided, skipping cleanup.")
        return 0

    cleanup_threshold = dt - timedelta(seconds=threshold_seconds)
    logger.debug("Cleaning up messages older than %s.", cleanup_threshold)

    num_deleted, _ = SlackMessage.objects.filter(created__lt=cleanup_threshold).delete()
    logger.info("Deleted %d old messages.", num_deleted)

    return num_deleted



from factory import Factory, SubFactory
from slack_sdk.errors import SlackApiError
from slack_sdk.web import SlackResponse


class SlackResponseFactory(Factory):
    client = None
    http_verb = "POST"
    api_url = ""
    req_args: dict = {  # noqa: RUF012
        "headers": {},
    }
    data: dict = {  # noqa: RUF012
        "ok": True,
    }
    headers: dict = {}  # noqa: RUF012
    status_code = 200

    class Meta:
        model = SlackResponse


class SlackApiErrorFactory(Factory):
    message = "Something went wrong"
    response = SubFactory(SlackResponseFactory, data={"ok": False})

    class Meta:
        model = SlackApiError



"""Utils for Slack messaging."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List, Optional

NoneType = type(None)


@dataclass
class MessageHeader:
    """Message header data definition."""

    mrkdwn: Optional[str] = field(default=None)  # noqa: UP007
    parse: Optional[str] = field(default=None)  # noqa: UP007
    reply_broadcast: Optional[bool] = field(default=None)  # noqa: UP007
    thread_ts: Optional[str] = field(default=None)  # noqa: UP007
    unfurl_links: Optional[bool] = field(default=None)  # noqa: UP007
    unfurl_media: Optional[bool] = field(default=None)  # noqa: UP007

    def __post_init__(self) -> None:
        _assert_type(self.mrkdwn, (str, NoneType))
        _assert_type(self.parse, (str, NoneType))
        _assert_type(self.reply_broadcast, (bool, NoneType))
        _assert_type(self.thread_ts, (str, NoneType))
        _assert_type(self.unfurl_links, (bool, NoneType))
        _assert_type(self.unfurl_media, (bool, NoneType))

    @classmethod
    def from_any(
        cls,
        obj: MessageHeader | dict[str, Any] | None = None,
    ) -> MessageHeader:
        """Create instance from compatible types."""
        if obj is None:
            return cls()

        if isinstance(obj, dict):
            return cls(**obj)

        msg = f"Unsupported type {type(obj)}"
        raise TypeError(msg)


@dataclass
class MessageBody:
    """Data definition for message body."""

    attachments: Optional[List[dict]] = field(default=None)  # noqa: UP006, UP007

    # See more about blocks at https://api.slack.com/reference/block-kit/blocks
    blocks: Optional[List[dict]] = field(default=None)  # noqa: UP006, UP007

    text: Optional[str] = field(default=None)  # noqa: UP007
    icon_emoji: Optional[str] = field(default=None)  # noqa: UP007
    icon_url: Optional[str] = field(default=None)  # noqa: UP007
    metadata: Optional[dict] = field(default=None)  # noqa: UP007
    username: Optional[str] = field(default=None)  # noqa: UP007

    def __post_init__(self) -> None:
        _assert_type(self.attachments, (list, NoneType))
        _assert_type(self.blocks, (list, NoneType))
        _assert_type(self.text, (str, NoneType))
        _assert_type(self.icon_emoji, (str, NoneType))
        _assert_type(self.icon_url, (str, NoneType))
        _assert_type(self.metadata, (dict, NoneType))
        _assert_type(self.username, (str, NoneType))

        if not any((self.attachments, self.blocks, self.text)):
            msg = "At least one of `attachments`, `blocks` and `text` must set"
            raise ValueError(msg)

    @classmethod
    def from_any(cls, obj: str | MessageBody | dict[str, Any]) -> MessageBody:
        """Create instance from compatible types."""
        if isinstance(obj, dict):
            return cls(**obj)

        if isinstance(obj, str):
            return cls(text=obj)

        msg = f"Unsupported type {type(obj)}"
        raise TypeError(msg)


def _assert_type(obj: Any, cls: type | tuple[type, ...]) -> None:
    if not isinstance(obj, cls):
        msg = f"Invalid value type, expected {cls}, got {type(obj)}"
        raise TypeError(msg)



from .mention import SlackMention
from .message import SlackMessage
from .message_recipient import SlackMessageRecipient
from .messaging_policy import SlackMessagingPolicy

__all__ = ("SlackMention", "SlackMessage", "SlackMessageRecipient", "SlackMessagingPolicy")



import pytest

from django_slack_tools.utils.slack import get_block_kit_builder_url


def test_get_block_kit_builder_url() -> None:
    with pytest.raises(ValueError, match="Only one of `blocks` or `attachments` should be provided."):
        get_block_kit_builder_url()

    with pytest.raises(ValueError, match="Only one of `blocks` or `attachments` should be provided."):
        get_block_kit_builder_url(blocks=[], attachments=[])

    payload = [
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": "Hello, World!",
            },
        },
    ]

    # Without team ID; just hoping the browser to take user right place (Slack may redirect with team ID if logged in)
    url = get_block_kit_builder_url(blocks=payload)
    assert (
        url
        == "https://app.slack.com/block-kit-builder/#%7B%22blocks%22%3A%20%5B%7B%22type%22%3A%20%22section%22%2C%20%22text%22%3A%20%7B%22type%22%3A%20%22mrkdwn%22%2C%20%22text%22%3A%20%22Hello%2C%20World%21%22%7D%7D%5D%7D"
    )

    url = get_block_kit_builder_url(attachments=payload)
    assert (
        url
        == "https://app.slack.com/block-kit-builder/#%7B%22attachments%22%3A%20%5B%7B%22type%22%3A%20%22section%22%2C%20%22text%22%3A%20%7B%22type%22%3A%20%22mrkdwn%22%2C%20%22text%22%3A%20%22Hello%2C%20World%21%22%7D%7D%5D%7D"
    )

    # With team ID
    url = get_block_kit_builder_url(team_id="T00000000", blocks=payload)
    assert (
        url
        == "https://app.slack.com/block-kit-builder/T00000000#%7B%22blocks%22%3A%20%5B%7B%22type%22%3A%20%22section%22%2C%20%22text%22%3A%20%7B%22type%22%3A%20%22mrkdwn%22%2C%20%22text%22%3A%20%22Hello%2C%20World%21%22%7D%7D%5D%7D"
    )

    url = get_block_kit_builder_url(team_id="T00000000", attachments=payload)
    assert (
        url
        == "https://app.slack.com/block-kit-builder/T00000000#%7B%22attachments%22%3A%20%5B%7B%22type%22%3A%20%22section%22%2C%20%22text%22%3A%20%7B%22type%22%3A%20%22mrkdwn%22%2C%20%22text%22%3A%20%22Hello%2C%20World%21%22%7D%7D%5D%7D"
    )



from __future__ import annotations

from typing import Any

import pytest

from django_slack_tools.slack_messages.models import SlackMessage
from tests._helpers import ModelTestBase

from ._factories import SlackMessageFactory


class TestSlackMessage(ModelTestBase):
    model_cls = SlackMessage
    factory_cls = SlackMessageFactory

    @pytest.mark.parametrize(
        argnames=["kwargs", "expect"],
        argvalues=[
            [{"ts": "1703393199.592689", "ok": True}, "Message (1703393199.592689, OK)"],
            [{"id": 1, "ok": False}, "Message (1, not OK)"],
            [{"id": 2, "ok": None}, "Message (2, not sent)"],
        ],
        ids=["ok", "not ok", "not sent"],
    )
    def test_str(self, kwargs: dict[str, Any], expect: str) -> None:
        instance = self.factory_cls.build(**kwargs)
        assert str(instance) == expect



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from todo.slack_app import app

from .models import Todo

if TYPE_CHECKING:
    from slack_bolt.context.ack import Ack
    from slack_bolt.context.say import Say
    from slack_sdk import WebClient

logger = logging.getLogger(__name__)


# App home
# https://api.slack.com/start/building/bolt-python
@app.event("app_home_opened")
def update_home_tab(client: WebClient, event: dict) -> None:
    """Show the user's To Do list when App Home is opened."""
    recent_todos = Todo.objects.filter(completed=False).order_by("-last_modified")[:5]  # type: ignore[attr-defined]
    recent_todo_element = {
        "type": "rich_text",
        "elements": [
            {
                "type": "rich_text_list",
                "style": "bullet",
                "elements": [
                    {
                        "type": "rich_text_section",
                        "elements": [
                            {
                                "type": "text",
                                "text": f"{todo.title}\n",
                                "style": {
                                    "bold": True,
                                },
                            },
                            {
                                "type": "text",
                                "text": "{description}\n".format(
                                    description=todo.description or "No description",
                                ),
                            },
                            {
                                "type": "text",
                                "text": f"Created in {todo.created:%Y-%m-%d}",
                            },
                        ],
                    }
                    for todo in recent_todos
                ],
            },
        ],
    }
    client.views_publish(
        user_id=event["user"],
        view={
            "type": "home",
            "callback_id": "home_view",
            "blocks": [
                {
                    "type": "header",
                    "text": {
                        "type": "plain_text",
                        "text": "Welcome to Django Slack Tools :tada:",
                    },
                },
                {"type": "divider"},
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "This is a sample Slack app built with <https://slack.dev/bolt-python/tutorial/getting-started|Bolt for Python> to display a list of To Do.",  # noqa: E501
                    },
                },
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": f"There is {len(recent_todos)} To Do not completed yet.",
                    },
                },
                recent_todo_element,
            ],
        },
    )


@app.shortcut("create_todo")
def create_todo(
    ack: Ack,
    client: WebClient,
    body: dict,
) -> None:
    """Shortcut to create a new To Do."""
    ack()
    client.views_open(
        trigger_id=body["trigger_id"],
        view={
            "type": "modal",
            "callback_id": "create_todo",
            "title": {"type": "plain_text", "text": "Create new To Do", "emoji": True},
            "submit": {"type": "plain_text", "text": "Submit", "emoji": True},
            "close": {"type": "plain_text", "text": "Cancel", "emoji": True},
            "blocks": [
                {
                    "block_id": "title",
                    "type": "input",
                    "element": {
                        "type": "plain_text_input",
                        "action_id": "plain_text_input-action",
                    },
                    "label": {"type": "plain_text", "text": "Title", "emoji": True},
                },
                {
                    "block_id": "description",
                    "type": "input",
                    "optional": True,
                    "element": {
                        "type": "plain_text_input",
                        "multiline": True,
                        "action_id": "plain_text_input-action",
                    },
                    "label": {
                        "type": "plain_text",
                        "text": "Description",
                        "emoji": True,
                    },
                },
            ],
        },
    )


@app.view("create_todo")
def handle_submission_create_todo(ack: Ack, body: dict) -> None:
    """Handle submission of a new To Do."""
    ack()
    values = body["view"]["state"]["values"]
    Todo.objects.create(  # type: ignore[attr-defined]
        title=values["title"]["plain_text_input-action"]["value"],
        description=values["description"]["plain_text_input-action"]["value"] or "",
    )


@app.shortcut("mark_todo_completed")
def mark_todo_completed(
    ack: Ack,
    client: WebClient,
    body: dict,
) -> None:
    """Shortcut to mark a To Do as completed."""
    logger.info("Received a body: %r", body)
    ack()
    incomplete_todos = Todo.objects.filter(completed=False).order_by("-last_modified")  # type: ignore[attr-defined]
    options = [
        {
            "text": {
                "type": "plain_text",
                "text": todo.title,
                "emoji": True,
            },
            "value": str(todo.id),
        }
        for todo in incomplete_todos
    ]
    client.views_open(
        trigger_id=body["trigger_id"],
        view={
            "type": "modal",
            "callback_id": "mark_todo_completed",
            "title": {"type": "plain_text", "text": "Create new To Do", "emoji": True},
            "submit": {"type": "plain_text", "text": "Submit", "emoji": True},
            "close": {"type": "plain_text", "text": "Cancel", "emoji": True},
            "blocks": [
                {
                    "block_id": "choice",
                    "type": "input",
                    "element": {
                        "type": "static_select",
                        "placeholder": {
                            "type": "plain_text",
                            "text": "Select an item",
                            "emoji": True,
                        },
                        "options": options,
                        "action_id": "static_select-action",
                    },
                    "label": {
                        "type": "plain_text",
                        "text": "Completed To Do",
                        "emoji": True,
                    },
                },
            ],
        },
    )


@app.view("mark_todo_completed")
def handle_submission_mark_todo_completed(ack: Ack, body: dict) -> None:
    """Handle submission of a new To Do."""
    ack()
    id_ = body["view"]["state"]["values"]["choice"]["static_select-action"]["selected_option"]["value"]
    todo = Todo.objects.get(id=id_)  # type: ignore[attr-defined]
    todo.completed = True
    todo.save()


@app.event("app_mention")
def handle_app_mention(event: dict, say: Say) -> None:
    """Reply greet to user."""
    logger.info("Received an event: %r", event)
    say(f"Hi there, <@{event['user']}>!")



"""Django form widgets."""

from __future__ import annotations

import json
from logging import getLogger

from django.forms import widgets

logger = getLogger(__name__)


# Ref from: https://stackoverflow.com/a/52627264
class JSONWidget(widgets.Textarea):
    """Pretty formatted JSON widget."""

    indent = 2

    def format_value(self, value: str) -> str | None:  # noqa: D102
        try:
            value = json.dumps(json.loads(value), indent=self.indent)
            lines = value.split("\n")
            width, height = max(len(ln) for ln in lines), len(lines)
            self.attrs["rows"] = min(max(height + self.indent, 10), 40)  # 10 ~ 40
            self.attrs["cols"] = min(max(width + self.indent, 40), 120)  # 40 ~ 120
        except Exception:
            logger.exception("Error while formatting JSON object")

        return super().format_value(value)



from __future__ import annotations

from typing import Any

from django.db.models import signals
from django.dispatch import receiver

from django_slack_tools.slack_messages.message import slack_message_via_policy

from .models import Todo


@receiver(signals.post_save, sender=Todo)
def notify_slack(
    sender: type[Todo],  # noqa: ARG001
    instance: Todo,
    created: bool,  # noqa: FBT001
    **kwargs: Any,  # noqa: ARG001
) -> None:
    """Send a Slack message if new To Do created or existing instance updated."""
    if created:
        slack_message_via_policy(
            "TODO-CREATED",
            lazy=True,
            context={"title": instance.title},
        )
    else:
        slack_message_via_policy(
            "TODO-UPDATED",
            lazy=True,
            context={"title": instance.title},
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from django.core.exceptions import ImproperlyConfigured
from slack_bolt import App

from django_slack_tools.app_settings import AppSettings
from django_slack_tools.slack_messages.backends import BaseBackend

if TYPE_CHECKING:
    from pytest_django.fixtures import SettingsWrapper

    from django_slack_tools.app_settings import ConfigDict


# Config fixtures to run in parametrize (map of test id to config dictionary)
config_fixtures: dict[str, ConfigDict] = {
    "dummy backend": {
        "SLACK_APP": "tests.test_app_settings.get_slack_app",
        "BACKEND": {
            "NAME": "django_slack_tools.slack_messages.backends.DummyBackend",
            "OPTIONS": {},
        },
    },
    "logging backend": {
        "SLACK_APP": "tests.test_app_settings.get_slack_app",
        "BACKEND": {
            "NAME": "django_slack_tools.slack_messages.backends.LoggingBackend",
            "OPTIONS": {},
        },
    },
    "slack backend": {
        "SLACK_APP": "tests.test_app_settings.get_slack_app",
        "BACKEND": {
            "NAME": "django_slack_tools.slack_messages.backends.SlackBackend",
            "OPTIONS": {
                "slack_app": "tests.test_app_settings.get_slack_app",
            },
        },
    },
    "slack redirect backend": {
        "SLACK_APP": "tests.test_app_settings.get_slack_app",
        "BACKEND": {
            "NAME": "django_slack_tools.slack_messages.backends.SlackRedirectBackend",
            "OPTIONS": {
                "slack_app": "tests.test_app_settings.get_slack_app",
                "redirect_channel": "some-redirect-channel",
            },
        },
    },
}

# Save decorator for reuse, as not all test in suite requires this
settings_dict_parametrizer = pytest.mark.parametrize(
    argnames="settings_dict",
    argvalues=config_fixtures.values(),
    ids=config_fixtures.keys(),
)


class TestAppSettings:
    def _assert_app_settings(self, app_settings: AppSettings) -> None:
        assert issubclass(app_settings.backend.__class__, BaseBackend)

    @settings_dict_parametrizer
    def test_dict_config(self, settings_dict: ConfigDict) -> None:
        app_settings = AppSettings(settings_dict)
        self._assert_app_settings(app_settings)

    @settings_dict_parametrizer
    def test_django_config(self, settings: SettingsWrapper, settings_dict: ConfigDict) -> None:
        settings.DJANGO_SLACK_TOOLS = None
        with pytest.raises(ImproperlyConfigured, match=r"^Neither `.+` provided or `.+` settings found"):
            AppSettings()

        settings.DJANGO_SLACK_TOOLS = settings_dict
        app_settings = AppSettings()
        self._assert_app_settings(app_settings)

    def test_raises_if_backend_is_not_subclass_of_base_class(self) -> None:
        with pytest.raises(
            ImproperlyConfigured,
            match="Provided backend is not a subclass of `django_slack_tools.slack_messages.backends.base.BaseBackend` class.",  # noqa: E501
        ):
            AppSettings(
                {
                    "SLACK_APP": "tests.test_app_settings.get_slack_app",
                    "BACKEND": {"NAME": "tests.test_app_settings.ImproperBackend", "OPTIONS": {}},
                },
            )

    def test_bad_config_not_slack_app(self) -> None:
        def not_slack_app() -> str:
            return "I'm not a slack app"

        with pytest.raises(ImproperlyConfigured, match="Couldn't resolve provided app spec into Slack app instance."):
            AppSettings(
                {
                    "SLACK_APP": "tests.test_app_settings.get_slack_app",
                    "BACKEND": {
                        "NAME": "django_slack_tools.slack_messages.backends.SlackBackend",
                        "OPTIONS": {
                            "slack_app": not_slack_app,
                        },
                    },
                },
            )

        with pytest.raises(ImproperlyConfigured, match="Provided `SLACK_APP` config is not Slack app."):
            AppSettings(
                {
                    "SLACK_APP": not_slack_app,  # type: ignore[typeddict-item]
                    "BACKEND": {
                        "NAME": "django_slack_tools.slack_messages.backends.SlackBackend",
                        "OPTIONS": {
                            "slack_app": "tests.test_app_settings.get_slack_app",
                        },
                    },
                },
            )


def get_slack_app() -> App:
    return App(token="i-am-a-cookie", token_verification_enabled=False)  # noqa: S106


class ImproperBackend:
    pass



from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest

from django_slack_tools.slack_messages import tasks
from django_slack_tools.slack_messages.models import SlackMessage
from tests.slack_messages.models._factories import SlackMessageFactory

pytestmark = pytest.mark.django_db


class TestSlackMessage:
    def test_slack_message(self) -> None:
        with mock.patch("django_slack_tools.slack_messages.message.slack_message") as m:
            tasks.slack_message(
                "Hello, world!",
                channel="test",
                header={"key": "value"},
                raise_exception=True,
                get_permalink=True,
            )

        m.assert_called_once_with(
            "Hello, world!",
            channel="test",
            header={"key": "value"},
            raise_exception=True,
            get_permalink=True,
        )


class TestSlackMessageViaPolicy:
    def test_slack_message_via_policy(self) -> None:
        with mock.patch("django_slack_tools.slack_messages.message.slack_message_via_policy") as m:
            tasks.slack_message_via_policy(
                "TEST-POLICY",
                header={"key": "value"},
                raise_exception=True,
                lazy=True,
                get_permalink=True,
                context={"another-key": "another-value"},
            )

        m.assert_called_once_with(
            "TEST-POLICY",
            header={"key": "value"},
            raise_exception=True,
            lazy=True,
            get_permalink=True,
            context={"another-key": "another-value"},
        )


class TestCleanupOldMessages:
    def test_cleanup_old_messages(self) -> None:
        # Arrange
        ts = datetime(2024, 10, 9, 3, 48, 22, tzinfo=timezone.utc)
        _should_deleted = [
            # Should be deleted
            SlackMessageFactory(created=ts - timedelta(minutes=6)),
            SlackMessageFactory(created=ts - timedelta(minutes=5, seconds=1)),
        ]
        should_remain = [
            # Should not be deleted
            SlackMessageFactory(created=ts - timedelta(minutes=5)),
            SlackMessageFactory(created=ts - timedelta(minutes=4, seconds=59)),
            SlackMessageFactory(created=ts - timedelta(minutes=4)),
        ]

        # Act
        num_deleted = tasks.cleanup_old_messages(base_ts=ts.isoformat(), threshold_seconds=5 * 60)  # 5 minutes

        # Assert
        assert num_deleted == 2
        assert sorted(SlackMessage.objects.values_list("id", flat=True)) == sorted(m.id for m in should_remain)

    def test_cleanup_old_messages_skip_if_threshold_is_none(self) -> None:
        # Arrange
        ts = datetime(2024, 10, 9, 3, 48, 22, tzinfo=timezone.utc)
        SlackMessageFactory(created=ts - timedelta(weeks=4))
        SlackMessageFactory(created=ts - timedelta(hours=1))
        SlackMessageFactory()

        # Act
        num_deleted = tasks.cleanup_old_messages(base_ts=ts.isoformat(), threshold_seconds=None)

        # Assert
        assert num_deleted == 0
        assert SlackMessage.objects.count() == 3



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable
from unittest import mock

import pytest

from django_slack_tools.slack_messages.admin import SlackMentionAdmin
from django_slack_tools.slack_messages.admin.mention import _get_mentionable_items
from django_slack_tools.slack_messages.models import SlackMention
from tests._factories import SlackResponseFactory
from tests._helpers import ModelAdminTestBase
from tests.slack_messages.models._factories import SlackMentionFactory

if TYPE_CHECKING:
    from django.test import Client
    from django.test.client import _MonkeyPatchedWSGIResponse

_mentionable_items = {
    "USER001": {"type": SlackMention.MentionType.USER, "name": "Cake"},
    "USER002": {"type": SlackMention.MentionType.USER, "name": "Carrot"},
    "GROUP01": {"type": SlackMention.MentionType.GROUP, "name": "Food"},
}


class TestSlackMentionAdmin(ModelAdminTestBase):
    admin_cls = SlackMentionAdmin
    model_cls = SlackMention
    factory_cls = SlackMentionFactory

    pytestmark = pytest.mark.django_db()

    def _update_mentions(
        self,
        *,
        client: Client,
        ids: Iterable[SlackMention],
    ) -> _MonkeyPatchedWSGIResponse:
        return client.post(
            self._reverse("changelist"),
            {
                "action": "_update_mentions",
                "_selected_action": ids,
            },
        )

    def test_update_mentions(self, admin_client: Client) -> None:
        # These mentions should update
        mentions_to_update = [
            self.factory_cls.create(mention_id="USER001"),
            self.factory_cls.create(mention_id="USER002"),
            self.factory_cls.create(mention_id="GROUP01"),
        ]
        ids = [m.id for m in mentions_to_update]

        # Perform action
        with mock.patch(
            "django_slack_tools.slack_messages.admin.mention._get_mentionable_items",
            return_value=_mentionable_items,
        ):
            response = self._update_mentions(client=admin_client, ids=ids)

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == ["Updated 3 mentions."]

        # Should update
        mentions = SlackMention.objects.filter(id__in=ids)
        assert list(mentions.values("type", "name")) == [
            {"type": SlackMention.MentionType.USER, "name": "Cake"},
            {"type": SlackMention.MentionType.USER, "name": "Carrot"},
            {"type": SlackMention.MentionType.GROUP, "name": "Food"},
        ]

    def test_update_mentions_no_match_for_some(self, admin_client: Client) -> None:
        # These mentions should update
        mentions_to_update = [
            self.factory_cls.create(mention_id="USER001"),
            self.factory_cls.create(mention_id="GROUP01"),
            # No match for this, so it should leave unchanged
            self.factory_cls.create(type=SlackMention.MentionType.UNKNOWN, mention_id="GROUP02", name="Olive"),
        ]
        ids = [m.id for m in mentions_to_update]

        # Perform action
        with mock.patch(
            "django_slack_tools.slack_messages.admin.mention._get_mentionable_items",
            return_value=_mentionable_items,
        ):
            response = self._update_mentions(client=admin_client, ids=ids)

        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == [
            "Updated 2 mentions successfully and there were 1 mentions failed to update because no matching data.",
        ]

        # Should update
        mentions = SlackMention.objects.filter(id__in=ids)
        assert list(mentions.values("type", "name")) == [
            {"type": SlackMention.MentionType.USER, "name": "Cake"},
            {"type": SlackMention.MentionType.GROUP, "name": "Food"},
            # Should leave unchanged
            {"type": SlackMention.MentionType.UNKNOWN, "name": "Olive"},
        ]

    def test_update_mentions_no_match(self, admin_client: Client) -> None:
        # These mentions should update
        mentions_to_update = [
            self.factory_cls.create(type=SlackMention.MentionType.UNKNOWN, mention_id="🐳"),
            self.factory_cls.create(type=SlackMention.MentionType.UNKNOWN, mention_id="🤖"),
        ]
        ids = [m.id for m in mentions_to_update]

        # Perform action
        with mock.patch(
            "django_slack_tools.slack_messages.admin.mention._get_mentionable_items",
            return_value=_mentionable_items,
        ):
            response = self._update_mentions(client=admin_client, ids=ids)

        assert response.status_code == 302

        # Check message
        messages = self._get_messages(response.wsgi_request)
        assert messages == [
            "Updated 0 mentions successfully and there were 2 mentions failed to update because no matching data.",
        ]

        # Should update
        mentions = SlackMention.objects.filter(id__in=ids)
        assert list(mentions.values("type", "mention_id")) == [
            {"type": SlackMention.MentionType.UNKNOWN, "mention_id": "🐳"},
            {"type": SlackMention.MentionType.UNKNOWN, "mention_id": "🤖"},
        ]


def test_get_mentionable_items(mock_slack_client: mock.Mock) -> None:
    mock_slack_client.users_list.return_value = SlackResponseFactory(
        data={
            "ok": True,
            "members": [
                {"id": "USER001", "profile": {"display_name": "Cake", "real_name": ""}},
                {"id": "USER002", "profile": {"display_name": "", "real_name": "Carrot"}},
            ],
        },
    )
    mock_slack_client.usergroups_list.return_value = SlackResponseFactory(
        data={
            "ok": True,
            "usergroups": [{"id": "GROUP01", "name": "Food"}, {"id": "GROUP02", "name": "Drink"}],
        },
    )
    mentionable_items = _get_mentionable_items()

    assert mentionable_items == {
        "USER001": {
            "type": SlackMention.MentionType.USER,
            "name": "Cake",
        },
        "USER002": {
            "type": SlackMention.MentionType.USER,
            "name": "Carrot",
        },
        "GROUP01": {
            "type": SlackMention.MentionType.GROUP,
            "name": "Food",
        },
        "GROUP02": {
            "type": SlackMention.MentionType.GROUP,
            "name": "Drink",
        },
    }



"""Message model."""

from __future__ import annotations

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_slack_tools.utils.model_mixins import TimestampMixin
from django_slack_tools.utils.slack import body_validator, header_validator

from .messaging_policy import SlackMessagingPolicy


class SlackMessageManager(models.Manager["SlackMessage"]):
    """Manager for Slack messages."""


class SlackMessage(TimestampMixin, models.Model):
    """An Slack message."""

    policy = models.ForeignKey(
        SlackMessagingPolicy,
        verbose_name=_("Messaging Policy"),
        help_text=_("Messaging policy for this message."),
        null=True,  # Message can be built from scratch without using templates
        blank=True,
        on_delete=models.SET_NULL,
    )
    channel = models.CharField(
        verbose_name=_("Channel"),
        help_text=_("ID of channel this message sent to."),
        blank=False,
        max_length=128,  # Maximum length of channel name is 80 characters
    )
    header = models.JSONField(
        verbose_name=_("Header"),
        help_text=_(
            "Slack control arguments."
            " Allowed fields are `mrkdwn`, `parse`, `reply_broadcast`, `thread_ts`, `unfurl_links`, `unfurl_media`.",
        ),
        validators=[header_validator],
    )
    body = models.JSONField(
        verbose_name=_("Body"),
        help_text=_(
            "Message body."
            " Allowed fields are `attachments`, `body`, `text`, `icon_emoji`, `icon_url`, `metadata`, `username`.",
        ),
        validators=[body_validator],
    )
    ok = models.BooleanField(
        verbose_name=_("OK"),
        help_text=_("Whether Slack API respond with OK. If never sent, will be `null`."),
        null=True,
        default=None,
    )
    permalink = models.CharField(
        verbose_name=_("Permalink"),
        help_text=_("Permanent link for this message."),
        max_length=256,
        default="",
        blank=True,
    )

    # As ID, `ts` assigned by Slack, it is known after received response
    # By known, `ts` refers to timestamp (Format of `datetime.timestamp()`, e.g. `"1702737142.945359"`)
    ts = models.CharField(
        verbose_name=_("Message ID"),
        help_text=_("ID of an Slack message."),
        max_length=32,
        null=True,
        blank=True,
        unique=True,
    )
    parent_ts = models.CharField(
        verbose_name=_("Thread ID"),
        help_text=_("ID of current conversation thread."),
        max_length=32,
        default="",
        blank=True,
    )

    # Extraneous call detail for debugging
    request = models.JSONField(
        verbose_name=_("Request"),
        help_text=_("Dump of request content for debugging."),
        null=True,
        blank=True,
    )
    response = models.JSONField(
        verbose_name=_("Response"),
        help_text=_("Dump of response content for debugging."),
        null=True,
        blank=True,
    )
    exception = models.TextField(
        verbose_name=_("Exception"),
        help_text=_("Exception message if any."),
        blank=True,
    )

    objects: SlackMessageManager = SlackMessageManager()

    class Meta:  # noqa: D106
        verbose_name = _("Message")
        verbose_name_plural = _("Messages")
        ordering = ("-created",)

    def __str__(self) -> str:
        if self.ok is True:
            return _("Message ({ts}, OK)").format(id=self.id, ts=self.ts)

        if self.ok is False:
            return _("Message ({id}, not OK)").format(id=self.id)

        return _("Message ({id}, not sent)").format(id=self.id)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from django_slack_tools.slack_messages.message import slack_message, slack_message_via_policy
from django_slack_tools.slack_messages.models import SlackMention, SlackMessage, SlackMessagingPolicy
from tests.slack_messages.models._factories import (
    SlackMentionFactory,
    SlackMessageRecipientFactory,
    SlackMessagingPolicyFactory,
)

from ._factories import SlackMessageResponseFactory

if TYPE_CHECKING:
    from unittest.mock import Mock

pytestmark = pytest.mark.django_db


def test_slack_message(mock_slack_client: Mock) -> None:
    mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()
    msg = slack_message("Hello, World!", channel="whatever-channel")

    assert isinstance(msg, SlackMessage)
    assert SlackMessage.objects.filter(id=msg.id).exists()
    assert msg.policy is None
    assert msg.channel == "whatever-channel"
    assert msg.header == {
        "mrkdwn": None,
        "parse": None,
        "reply_broadcast": None,
        "thread_ts": None,
        "unfurl_links": None,
        "unfurl_media": None,
    }
    assert msg.body["text"] == "Hello, World!"
    assert msg.ok
    assert msg.permalink == ""
    assert msg.ts
    assert msg.parent_ts == ""
    assert isinstance(msg.request, dict)
    assert isinstance(msg.response, dict)
    assert msg.exception == ""


def test_slack_message_via_policy_dict_template(mock_slack_client: Mock) -> None:
    recipients = [
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
        SlackMessageRecipientFactory(
            mentions=[SlackMentionFactory(type=SlackMention.MentionType.SPECIAL, mention_id="<!here>")],
        ),
    ]
    policy = SlackMessagingPolicyFactory(
        code="TEST-PO-001",
        template={
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "{greet}, {mentions}",
                    },
                },
            ],
        },
        recipients=recipients,
    )
    mock_slack_client.chat_postMessage.side_effect = SlackMessageResponseFactory.create_batch(size=3)
    num_sent = slack_message_via_policy(policy, context={"greet": "Nice to meet you"})

    assert num_sent == 3
    assert SlackMessage.objects.filter(policy=policy).count() == 3


def test_slack_message_via_policy_django_template(mock_slack_client: Mock) -> None:
    recipients = [
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
        SlackMessageRecipientFactory(
            mentions=[SlackMentionFactory(type=SlackMention.MentionType.SPECIAL, mention_id="<!here>")],
        ),
    ]
    policy = SlackMessagingPolicyFactory(
        code="TEST-PO-001-DJ",
        template_type=SlackMessagingPolicy.TemplateType.DJANGO,
        template="greet.xml",
        recipients=recipients,
    )
    mock_slack_client.chat_postMessage.side_effect = SlackMessageResponseFactory.create_batch(size=3)
    num_sent = slack_message_via_policy(policy, context={"greet": "Nice to meet you"})

    assert num_sent == 3
    assert SlackMessage.objects.filter(policy=policy).count() == 3


def test_slack_message_via_policy_django_inline_template(mock_slack_client: Mock) -> None:
    recipients = [
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
        SlackMessageRecipientFactory(
            mentions=[SlackMentionFactory(type=SlackMention.MentionType.SPECIAL, mention_id="<!here>")],
        ),
    ]
    policy = SlackMessagingPolicyFactory(
        code="TEST-PO-001-DJI",
        template_type=SlackMessagingPolicy.TemplateType.DJANGO_INLINE,
        template="""
            <root>
                <block type="section">
                    <text type="mrkdwn">
                        {{ greet }}, {{ mentions }}
                    </text>
                </block>
            </root>
        """,
        recipients=recipients,
    )
    mock_slack_client.chat_postMessage.side_effect = SlackMessageResponseFactory.create_batch(size=3)
    num_sent = slack_message_via_policy(policy, context={"greet": "Nice to meet you"})

    assert num_sent == 3
    assert SlackMessage.objects.filter(policy=policy).count() == 3


def test_slack_message_via_policy_unknown_template_type(mock_slack_client: Mock) -> None:
    recipients = [
        SlackMessageRecipientFactory(mentions=SlackMentionFactory.create_batch(size=2)),
    ]
    policy = SlackMessagingPolicyFactory(
        code="TEST-PO-001-?",
        template_type=SlackMessagingPolicy.TemplateType.UNKNOWN,
        template=None,
        recipients=recipients,
    )
    mock_slack_client.chat_postMessage.side_effect = [SlackMessageResponseFactory()]
    with pytest.raises(ValueError, match="Unsupported template type: SlackMessagingPolicy.TemplateType.UNKNOWN"):
        slack_message_via_policy(policy, context={"greet": "Nice to meet you"})


def test_slack_message_via_policy_default(mock_slack_client: Mock) -> None:
    default_policy = SlackMessagingPolicy.objects.get(code="DEFAULT")
    default_policy.enabled = True
    default_policy.template = {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": "{greet}, {mentions_as_str}",
                },
            },
        ],
    }
    default_policy.save()
    default_policy.recipients.add(
        SlackMessageRecipientFactory(
            channel="whatever-channel",
            mentions=[SlackMentionFactory(type=SlackMention.MentionType.SPECIAL, mention_id="<!channel>")],
        ),
    )

    mock_slack_client.chat_postMessage.side_effect = SlackMessageResponseFactory()
    num_sent = slack_message_via_policy(context={"greet": "Nice to meet you"})

    assert num_sent == 1
    assert SlackMessage.objects.filter(policy=default_policy).count() == 1


def test_slack_message_via_policy_policy_not_enabled(mock_slack_client: Mock) -> None:
    policy = SlackMessagingPolicyFactory(
        code="TEST-PO-002",
        enabled=False,
        recipients=[
            SlackMessageRecipientFactory(),
        ],
    )
    num_sent = slack_message_via_policy(policy.code, context={"greet": "Nice to meet you"})
    mock_slack_client.chat_postMessage.assert_not_called()

    assert num_sent == 0


def test_slack_message_via_policy_context_shadowing_defaults(mock_slack_client: Mock) -> None:
    """Test template kwargs being shadowed by user provided context."""
    policy = SlackMessagingPolicyFactory(code="TEST", recipients=[])

    # As there is no recipient, no message will be sent
    # It's just OK no exception being thrown
    num_sent = slack_message_via_policy(policy.code, context={"mentions_as_str": "💣"})
    mock_slack_client.chat_postMessage.assert_not_called()

    assert num_sent == 0


def test_slack_message_via_policy_lazy(mock_slack_client: Mock) -> None:
    # Policy not exist at first
    code = "TEST-PO-LAZY-001"
    assert not SlackMessagingPolicy.objects.filter(code=code).exists()

    # Make call with lazy mode
    num_sent = slack_message_via_policy(code, lazy=True, context={"message": "Nice to meet you"})
    mock_slack_client.chat_postMessage.assert_not_called()

    # Ensure policy has been created
    policy = SlackMessagingPolicy.objects.get(code=code)
    assert policy.code == code
    assert policy.enabled is False
    assert list(policy.recipients.values_list("alias", flat=True)) == ["DEFAULT"]
    assert policy.template == {"text": "No template configured for lazily created policy {policy}"}

    # No message will be sent
    assert num_sent == 0


def test_slack_message_via_policy_lazy_already_exists(mock_slack_client: Mock) -> None:
    policy = SlackMessagingPolicyFactory(
        code="TEST-PO-LAZY-002",
        enabled=True,
        recipients=[
            SlackMessageRecipientFactory(),
        ],
    )

    # Make call with lazy mode
    mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()
    num_sent = slack_message_via_policy(policy.code, lazy=True, context={"message": "Nice to meet you"})

    # No message will be sent
    assert num_sent == 1



from __future__ import annotations

import json
from pathlib import Path

import pytest
from django.template import TemplateDoesNotExist

from django_slack_tools.utils.template.django import DjangoTemplate


@pytest.fixture(scope="session")
def data_dir() -> Path:
    return Path(__file__).parent / "testcases"


class TestDjangoTemplate:
    def test_instance_creation(self) -> None:
        # Mutually exclusive arguments provided
        with pytest.raises(TypeError, match="Exactly one of 'file' or 'inline' must be provided."):
            DjangoTemplate(file="whatever.xml", inline="<whatever>Whatever</whatever>")  # type: ignore[call-overload]

        # Unsupported file extension
        with pytest.raises(TemplateDoesNotExist, match="template-does-not-exists.xml"):
            DjangoTemplate(file="template-does-not-exists.xml")

    @pytest.mark.parametrize(
        ("xml_input", "json_expect"),
        [
            ("001.xml", "001.json"),
        ],
    )
    def test_render(self, xml_input: str, json_expect: str, data_dir: Path) -> None:
        expect = json.loads((data_dir / json_expect).read_text())
        actual = DjangoTemplate(inline=(data_dir / xml_input).read_text()).render()
        assert actual == expect

    def test_render_complex_with_context(self, data_dir: Path) -> None:
        # Arrange
        template = DjangoTemplate(file="001-template.xml")
        context = {
            "restaurants": [
                {
                    "id": "click_me_123",
                    "title": "Farmhouse Thai Cuisine",
                    "rating": 4,
                    "reviews": 1528,
                    "comment": (
                        "They do have some vegan options, like the roti and curry, plus they have a ton of"
                        " salad stuff and noodles can be ordered without meat!! They have something for"
                        " everyone here"
                    ),
                    "image_url": "https://s3-media3.fl.yelpcdn.com/bphoto/c7ed05m9lC2EmA3Aruue7A/o.jpg",
                    "action": {
                        "id": "click_me_123",
                        "display": "Farmhouse",
                        "url": "",
                        "emoji": False,
                    },
                },
                {
                    "id": "click_me_123",
                    "title": "Kin Khao",
                    "rating": 4,
                    "reviews": 1638,
                    "comment": (
                        "The sticky rice also goes wonderfully with the caramelized pork belly, which is"
                        " absolutely melt-in-your-mouth and so soft."
                    ),
                    "image_url": "https://s3-media2.fl.yelpcdn.com/bphoto/korel-1YjNtFtJlMTaC26A/o.jpg",
                    "action": {
                        "id": "click_me_123",
                        "display": "Kin Khao",
                        "url": "https://google.com",
                        "emoji": True,
                    },
                },
                {
                    "title": "Ler Ros",
                    "rating": 4,
                    "reviews": 2082,
                    "comment": (
                        "I would really recommend the Yum Koh Moo Yang - Spicy lime dressing and roasted"
                        " quick marinated pork shoulder, basil leaves, chili & rice powder."
                    ),
                    "image_url": "https://s3-media2.fl.yelpcdn.com/bphoto/DawwNigKJ2ckPeDeDM7jAg/o.jpg",
                    "action": {
                        "id": "click_me_123",
                        "display": "Ler Ros",
                        "url": "https://google.com",
                        "emoji": True,
                    },
                },
            ],
        }

        # Act
        actual = template.render(context=context)

        # Assert
        expect = json.loads((data_dir / "001.json").read_text())
        assert actual == expect



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from slack_sdk.errors import SlackApiError

from django_slack_tools.slack_messages.backends import SlackBackend, SlackRedirectBackend
from django_slack_tools.slack_messages.models import SlackMessage
from django_slack_tools.utils.slack import MessageBody, MessageHeader
from tests._factories import SlackApiErrorFactory, SlackResponseFactory
from tests.slack_messages._factories import SlackMessageResponseFactory

if TYPE_CHECKING:
    from unittest.mock import Mock

    from slack_bolt import App


class TestSlackBackend:
    # TODO(lasuillard): Test `.__init__()` for import string & callable

    pytestmark = pytest.mark.django_db()

    @pytest.fixture
    def backend(self, slack_app: App) -> SlackBackend:
        return SlackBackend(slack_app=slack_app)

    def test_send_message(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        """Test sending message."""
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

        message = backend.send_message(
            backend.prepare_message(
                channel="test-channel",
                header=MessageHeader(),
                body=MessageBody(text="Hello, World!"),
            ),
            raise_exception=True,
            get_permalink=False,
        )

        assert isinstance(message, SlackMessage)
        assert message.request
        assert "Authorization" not in message.request["headers"]
        assert message.response
        assert not message.exception

    def test_send_message_error_response(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        """Test `raise_exception` flag."""
        message = SlackMessage(channel="test-channel", header={}, body={})
        mock_slack_client.chat_postMessage.side_effect = SlackApiErrorFactory()

        # Should re-raise the exception
        with pytest.raises(SlackApiError):
            backend.send_message(message, raise_exception=True, get_permalink=False)

        # Won't raise
        backend.send_message(message, raise_exception=False, get_permalink=False)

    def test_send_message_unexpected_error(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        """Test `raise_exception` flag."""
        message = SlackMessage(channel="test-channel", header={}, body={})

        class UnexpectedError(Exception): ...

        mock_slack_client.chat_postMessage.side_effect = UnexpectedError()

        # Should re-raise the exception
        with pytest.raises(UnexpectedError):
            backend.send_message(message, raise_exception=True, get_permalink=False)

        # Won't raise
        backend.send_message(message, raise_exception=False, get_permalink=False)

    def test_send_message_permalink(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        """Test permalink."""
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()
        mock_slack_client.chat_getPermalink.return_value = SlackResponseFactory(
            data={"ok": True, "permalink": "https://..."},
        )

        message = backend.send_message(
            message=SlackMessage(channel="test-channel", header={}, body={}),
            raise_exception=False,
            get_permalink=True,
        )

        assert isinstance(message, SlackMessage)
        assert message.permalink == "https://..."

    def test_get_permalink(self, backend: SlackBackend, mock_slack_client: Mock) -> None:
        message = SlackMessage(channel="test-channel", ts="0000.0000")
        mock_slack_client.chat_getPermalink.return_value = SlackResponseFactory(
            data={"ok": True, "permalink": "https://..."},
        )
        permalink = backend._get_permalink(message=message, raise_exception=True)
        assert permalink == "https://..."

        # If error, returns empty string
        mock_slack_client.chat_getPermalink.side_effect = SlackApiErrorFactory()
        permalink = backend._get_permalink(message=message, raise_exception=False)
        assert permalink == ""

        # Re-raise exception if flag set
        mock_slack_client.chat_getPermalink.side_effect = SlackApiErrorFactory()
        with pytest.raises(SlackApiError):
            backend._get_permalink(message=message, raise_exception=True)

        # Raise exception if message timestamp is not set
        with pytest.raises(ValueError, match="Message timestamp is not set, can't retrieve permalink."):
            backend._get_permalink(message=SlackMessage(channel="test-channel", ts=None), raise_exception=True)


class TestSlackRedirectBackend:
    pytestmark = pytest.mark.django_db()

    @pytest.fixture
    def backend(self, slack_app: App) -> SlackRedirectBackend:
        return SlackRedirectBackend(slack_app=slack_app, redirect_channel="test-redirect-channel")

    def test_send_message(self, backend: SlackRedirectBackend, mock_slack_client: Mock) -> None:
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

        message = backend.send_message(
            backend.prepare_message(
                channel="whatever-this-channel",
                header=MessageHeader(),
                body=MessageBody(text="Hello, World!"),
            ),
            raise_exception=True,
            get_permalink=False,
        )

        assert isinstance(message, SlackMessage)

    def test_send_message_no_redirect(self, backend: SlackRedirectBackend, mock_slack_client: Mock) -> None:
        backend.inform_redirect = False
        mock_slack_client.chat_postMessage.return_value = SlackMessageResponseFactory()

        message = backend.send_message(
            backend.prepare_message(
                channel="whatever-this-channel",
                header=MessageHeader(),
                body=MessageBody(text="Hello, World!"),
            ),
            raise_exception=True,
            get_permalink=False,
        )

        assert message.body == {
            "text": "Hello, World!",
            "attachments": None,
            "blocks": None,
            "icon_emoji": None,
            "icon_url": None,
            "metadata": None,
            "username": None,
        }

    def test_prepare_message(self, backend: SlackRedirectBackend) -> None:
        message = backend.prepare_message(
            channel="test-original-channel",
            header=MessageHeader(),
            body=MessageBody(
                attachments=[{"text": "Django Slack Tools"}],
            ),
        )
        assert message.body == {
            "text": None,
            "attachments": [
                {
                    "color": "#eb4034",
                    "text": ":warning:  This message was originally sent to channel *test-original-channel* but redirected here.",  # noqa: E501
                },
                {"text": "Django Slack Tools"},
            ],
            "blocks": None,
            "icon_emoji": None,
            "icon_url": None,
            "metadata": None,
            "username": None,
        }

    def test_make_inform_attachment(self, backend: SlackRedirectBackend) -> None:
        attachment = backend._make_inform_attachment(original_channel="test-original-channel")
        assert attachment == {
            "color": "#eb4034",
            "text": ":warning:  This message was originally sent to channel *test-original-channel* but redirected here.",  # noqa: E501
        }



# noqa: D100
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from django.contrib import admin, messages
from django.contrib.admin.filters import DateFieldListFilter
from django.db.models import Count
from django.db.models.query import QuerySet
from django.utils.translation import gettext_lazy as _

from django_slack_tools.app_settings import app_settings
from django_slack_tools.slack_messages.models import SlackMessageRecipient

if TYPE_CHECKING:
    from django.http import HttpRequest

    class SlackMessageRecipientWithAnnotates(SlackMessageRecipient):  # noqa: D101
        num_mentions: int


@admin.register(SlackMessageRecipient)
class SlackMessageRecipientAdmin(admin.ModelAdmin):
    """Admin for recipients."""

    def get_queryset(self, request: HttpRequest) -> QuerySet[SlackMessageRecipientWithAnnotates]:  # noqa: D102
        return cast(
            QuerySet["SlackMessageRecipientWithAnnotates"],  # Unsafe force type casting
            super()
            .get_queryset(request)
            .annotate(
                # Avoid calling `.recipients.count()` per records
                num_mentions=Count("mentions"),
            ),
        )

    readonly_fields = ("id", "created", "last_modified")

    # Actions
    actions = ("_update_channel_names",)

    @admin.action(description=_("Update recipient' channel names"))
    def _update_channel_names(self, request: HttpRequest, queryset: QuerySet[SlackMessageRecipient]) -> None:
        """Admin action to update selected recipients' channel names using workspace info."""
        channels = _get_channels()

        # Temporary changes
        changes: list[SlackMessageRecipient] = []
        failures: list[SlackMessageRecipient] = []
        for recipient in queryset:
            match = channels.get(recipient.channel)
            if match is None:
                failures.append(recipient)
                continue

            recipient.channel_name = f"#{match}"
            changes.append(recipient)

        # Bulk update in single query
        SlackMessageRecipient.objects.bulk_update(changes, fields=("channel_name",))
        n_success = len(changes)

        # Reporting
        if failures:
            messages.warning(
                request,
                _(
                    "Updated {n_success} recipients successfully"
                    " and there were {n_fail} recipients failed to update because no matching data.",
                ).format(n_success=n_success, n_fail=len(failures)),
            )
        else:
            messages.info(request, _("Updated {n} recipients.").format(n=n_success))

    # Changelist
    # ------------------------------------------------------------------------
    date_hierarchy = "last_modified"
    search_fields = ("alias", "channel_name", "channel", "mentions__mention")
    list_display = ("id", "alias", "channel_name", "_num_mentions", "created", "last_modified")
    list_display_links = ("id", "alias", "channel_name")
    list_filter = (
        ("created", DateFieldListFilter),
        ("last_modified", DateFieldListFilter),
    )

    @admin.display(description=_("Number of Mentions"))
    def _num_mentions(self, instance: SlackMessageRecipientWithAnnotates) -> int:
        return instance.num_mentions

    # Change
    # ------------------------------------------------------------------------
    fieldsets = (
        (
            None,
            {
                "fields": ("alias", "channel", "channel_name", "mentions"),
            },
        ),
        (
            _("Miscellaneous"),
            {
                "fields": ("id", "created", "last_modified"),
                "classes": ("collapse",),
            },
        ),
    )


def _get_channels() -> dict[str, str]:
    # TODO(lasuillard): Need pagination in future
    response = app_settings.slack_app.client.conversations_list()
    channels: list[dict] = response.get("channels", default=[])
    return {channel["id"]: channel["name"] for channel in channels}



"""Slack backends actually interact with Slack API to do something."""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any, Callable

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from slack_bolt import App
from slack_sdk.errors import SlackApiError

from .base import BaseBackend

if TYPE_CHECKING:
    from slack_sdk.web import SlackResponse

    from django_slack_tools.slack_messages.models import SlackMessage
    from django_slack_tools.utils.slack import MessageBody


logger = getLogger(__name__)


class SlackBackend(BaseBackend):
    """Backend actually sending the messages."""

    def __init__(
        self,
        *,
        slack_app: App | Callable[[], App] | str,
    ) -> None:
        """Initialize backend.

        Args:
            slack_app: Slack app instance or import string.
        """
        if isinstance(slack_app, str):
            slack_app = import_string(slack_app)

        if callable(slack_app):
            slack_app = slack_app()

        if not isinstance(slack_app, App):
            msg = "Couldn't resolve provided app spec into Slack app instance."
            raise ImproperlyConfigured(msg)

        self._slack_app = slack_app

    def _send_message(self, message: SlackMessage) -> SlackResponse:
        header = message.header or {}
        body = message.body or {}
        return self._slack_app.client.chat_postMessage(channel=message.channel, **header, **body)

    def _get_permalink(self, *, message: SlackMessage, raise_exception: bool = False) -> str:
        """Get a permalink for given message identifier."""
        if not message.ts:
            msg = "Message timestamp is not set, can't retrieve permalink."
            raise ValueError(msg)

        try:
            _permalink_resp = self._slack_app.client.chat_getPermalink(
                channel=message.channel,
                message_ts=message.ts,
            )
        except SlackApiError:
            if raise_exception:
                raise

            logger.exception(
                "Error occurred while sending retrieving message's permalink,"
                " but ignored as `raise_exception` not set.",
            )
            return ""

        return _permalink_resp.get("permalink", default="")

    def _record_request(self, response: SlackResponse) -> dict[str, Any]:
        # Remove auth header (token) from request before recording
        response.req_args.get("headers", {}).pop("Authorization", None)

        return response.req_args

    def _record_response(self, response: SlackResponse) -> dict[str, Any]:
        return {
            "http_verb": response.http_verb,
            "api_url": response.api_url,
            "status_code": response.status_code,
            "headers": response.headers,
            "data": response.data,
        }


class SlackRedirectBackend(SlackBackend):
    """Inherited Slack backend with redirection to specific channels."""

    def __init__(self, *, slack_app: App | str, redirect_channel: str, inform_redirect: bool = True) -> None:
        """Initialize backend.

        Args:
            slack_app: Slack app instance or import string.
            redirect_channel: Slack channel to redirect.
            inform_redirect: Whether to append an attachment informing that the message has been redirected.
                Defaults to `True`.
        """
        self.redirect_channel = redirect_channel
        self.inform_redirect = inform_redirect

        super().__init__(slack_app=slack_app)

    def prepare_message(self, *args: Any, channel: str, body: MessageBody, **kwargs: Any) -> SlackMessage:
        """Prepare message to send, with modified for redirection.

        Args:
            args: Positional arguments to pass to super method.
            channel: Original channel to send message.
            body: Message content.
            kwargs: Keyword arguments to pass to super method.

        Returns:
            Prepared message instance.
        """
        # Modify channel to force messages always sent to specific channel
        # Add an attachment that informing message has been redirected
        if self.inform_redirect:
            body.attachments = [
                self._make_inform_attachment(original_channel=channel),
                *(body.attachments or []),
            ]

        return super().prepare_message(*args, channel=self.redirect_channel, body=body, **kwargs)

    def _make_inform_attachment(self, *, original_channel: str) -> dict[str, Any]:
        msg_redirect_inform = _(
            ":warning:  This message was originally sent to channel *{channel}* but redirected here.",
        )

        return {
            "color": "#eb4034",
            "text": msg_redirect_inform.format(channel=original_channel),
        }
